#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#

#
# CoAP message format based on:
#	- draft-ietf-core-coap-18
#	- draft-ietf-core-block-14
#	- draft-ietf-core-observe-11
#

from	ttproto.core.data		import *
from	ttproto.core.typecheck	import *
from	ttproto.core.subtype		import SubtypeClass
from	ttproto.core.union		import UnionClass
from	ttproto.core.packet		import PacketValue
from	ttproto.core.lib.inet.meta	import *
from	ttproto.core.lib.inet.basics	import *
from	ttproto.core			import exceptions
from	ttproto.core.templates	import Range, Length
from	ttproto.core.primitive	import IntValue

import	ttproto.core.lib.inet.udp

from urllib.parse import quote_plus


__all__ = [
    'CoAP',
    'CoAPOption',
    'CoAPOptionAccept',
    'CoAPOptionBlock',
    'CoAPOptionBlock1',
    'CoAPOptionBlock2',
    'CoAPOptionContentFormat',
    'CoAPOptionETag',
    'CoAPOptionEmpty',
    'CoAPOptionIfMatch',
    'CoAPOptionIfNoneMatch',
    'CoAPOptionList',
    'CoAPOptionLocationPath',
    'CoAPOptionLocationQuery',
    'CoAPOptionMaxAge',
    'CoAPOptionObserve',
    'CoAPOptionProxyUri',
    'CoAPOptionProxyScheme',
    'CoAPOptionSize1',
    'CoAPOptionSize2',
    'CoAPOptionString',
    'CoAPOptionUInt',
    'CoAPOptionUriHost',
    'CoAPOptionUriPath',
    'CoAPOptionUriPort',
    'CoAPOptionUriQuery',
    'CoAPOptionOneM2MFrom',
    'CoAPOptionOneM2MRequestIdentifier', 
    'CoAPOptionOneM2MName',
    'CoAPOptionOneM2MNotificationURI',
    'CoAPOptionOneM2MResponseStatusCode',
    'CoAPOptionOneM2MGroupRequestIdentifier',
    'CoAPOptionOneM2MTY',
    'CoAPCode',
    'CoAPType',

]

class _CoAPOptionHdrInt (metaclass = SubtypeClass (Range (int, 0, 65804))):
    pass

class _CoAPUInt (IntValue):

        @typecheck
        def _build_message (self) -> is_flatvalue_binary:

            result = b""
            v = self

            while v > 0:
                result = bytes((v & 0xff,)) + result
                v >>= 8

            return self, result

        @classmethod
        @typecheck
        def _decode_message (cls, bin_slice: BinarySlice) -> is_flatvalue_binslice:
            result = 0

            for c in bin_slice:
                result = result << 8 | c


            return cls(result), bin_slice[len (bin_slice):]


"""
     0   1   2   3   4   5   6   7
   +---------------+---------------+
   |               |               |
   |  Option Delta | Option Length |   1 byte
   |               |               |
   +---------------+---------------+
   \                               \
   /         Option Delta          /   0-2 bytes
   \          (extended)           \
   +-------------------------------+
   \                               \
   /         Option Length         /   0-2 bytes
   \          (extended)           \
   +-------------------------------+
   \                               \
   /                               /
   \                               \
   /         Option Value          /   0 or more bytes
   \                               \
   /                               /
   \                               \
   +-------------------------------+

   Option Delta:
   Option Length:  4-bit unsigned integer.  A value between 0 and 12
      indicates the length of the Option Value, in bytes.  Three values
      are reserved for special constructs:

      13:  An 8-bit unsigned integer precedes the Option Value and
         indicates the Option Length minus 13.

      14:  A 16-bit unsigned integer in network byte order precedes the
         Option Value and indicates the Option Length minus 269.

      15:  Reserved for future use.  If the field is set to this value,
         it MUST be processed as a message format error.
"""

class CoAPOption (
    metaclass = InetPacketClass,
    fields    = [
        ("Delta",	"dlt",	_CoAPOptionHdrInt,	0),
        ("Length",	"len",	_CoAPOptionHdrInt,	0),
        ("Value",	"val",	bytes,		b""),
    ]):

    def __init__ (self, *k, **kw):
        if len(k) == 1:
            kw["val"] = k[0]
            k = ()
        super().__init__ (*k, **kw)

    def _encode_delta_length (self, delta, length):
        def encode_int (v):
            if v >= 0:
                if v <= 12:
                    return v, b""
                elif v <= 268:
                    v -= 13
                    return 13, bytes((v,))
                elif v <= 65804:
                    v -= 269
                    return 14, bytes(((v // 256), (v % 256)))

            raise Exception ("Invalid option length/delta: %d" % v)

        d, ext_d = encode_int (delta)
        l, ext_l = encode_int (length)

        dl = bytes ((d*16 + l,))

        return dl + ext_d + ext_l

    def _build_message (self):
        length = self["len"]

        # prepare the fields and values
        fields = list (self.fields())
        values = self._fill_default_values()

        # do not build body for CoAPOptionEmpty (nothing to build)
        if isinstance(self,CoAPOptionEmpty):
            body_bin = b''
        else:
            # assemble the body of the option
            values[2:], bins = zip(*(f.tag.build_message (v) for v,f in zip(values[2:], fields[2:])))
            body_bin = concatenate (bins)


        if not isinstance (body_bin, bytes):
            raise Exception ("Option value length is not a multiple of 8 bits")

        # compute the option length if unset
        if length is None:
            values[1] = len (body_bin)

        # encode the delta & length
        dl_bin = self._encode_delta_length(values[0], values[1])

        return type(self) (*values), (dl_bin+body_bin)

    @classmethod
    def _decode_body (cls, values, bin_slice):
        fields = list(cls.fields())[2:]
        try:
            for field in fields:
                v, bin_slice = field.tag.decode_message (field.type, bin_slice)
                values.append (v)

            return bin_slice
        except Exception as e:
            exceptions.push_location (e, cls, field.name)
            raise

    @classmethod
    def _decode_message (cls, bin_slice, previous_type = None):

        # decode the delta & length
        d = bin_slice[0] >> 4
        l = bin_slice[0] &  0xf
        bin_slice = bin_slice[1:]

        def decode_int (v):
            nonlocal bin_slice
            if v == 13:
                v = 13 + bin_slice[0]
                bin_slice = bin_slice[1:]
            elif v == 14:
                v = 269 + bin_slice[0]*256 + bin_slice[1]
                bin_slice = bin_slice[2:]
            return v

        delta  = decode_int (d)
        length = decode_int (l)

        if previous_type is not None:
            option_type = previous_type + delta

            cls = cls.get_variant_type (option_type)

        # decode the other fields
        remaining_slice = bin_slice[length:]
        bin_slice       = bin_slice[:length]

        values = [delta, length]

        bin_slice = cls._decode_body(values, bin_slice)
        if bin_slice:
            raise Exception ("CoAP Option not fully decoded: %d bytes left in the buffer" % len(bin_slice))

        # create the packet
        return cls (*values), remaining_slice


class CoAPOptionList (
    metaclass = InetOrderedListClass,
    content_type = CoAPOption):

    @classmethod
    def _decode_message (cls, bin_slice, count = None):

        values = []
        current = 0

        try:
            while bin_slice and bin_slice[0] != 0xff:
                opt, bin_slice = CoAPOption._decode_message (bin_slice, current)

                current += opt["Delta"]
                values.append (opt)

        except Exception as e:
            exceptions.push_location (e, cls, str(len(values)))
            raise

        return cls (values), bin_slice

    def _build_message (self):

        assert self.is_flat()

        # If option deltas are unset, then we fill them and reorder the options if necessary
        #
        if all (opt["Delta"] is None for opt in self):

            # group the options by type
            by_type = {}
            for opt in self:
                t = opt.get_variant_id()
                lst = by_type.get (t)
                if lst is None:
                    by_type[t] = [opt]
                else:
                    lst.append (opt)

            # fill the deltas and generate the option list
            type_list = list (by_type)
            type_list.sort()
            current = 0
            values = []
            for t in type_list:
                for opt in by_type[t]:
                    delta   = t - current
                    current = t

                    v = list(opt)
                    v[0] = delta
                    values.append (type(opt)(*v))

            # overwrite the current value
            self = type(self) (values)

        # call InetList._build_message()
        return super()._build_message()

class _CoAPCodeDescription:
    __known_codes = {
            # Null message
            0:	"Empty",

            # Methods
            1:	"GET",
            2:	"POST",
            3:	"PUT",
            4:	"DELETE",

            # Responses
            65:	"2.01 Created",
            66:	"2.02 Deleted",
            67:	"2.03 Valid",
            68:	"2.04 Changed",
            69:	"2.05 Content",
            128:	"4.00 Bad Request",
            129:	"4.01 Unauthorized",
            130:	"4.02 Bad Option",
            131:	"4.03 Forbidden",
            132:	"4.04 Not Found",
            133:	"4.05 Method Not Allowed",
            134:	"4.06 Not Acceptable",
            140:	"4.12 Precondition Failed",
            141:	"4.13 Request Entity Too Large",
            143:	"4.15 Unsupported Media Type",
            160:	"5.00 Internal Server Error",
            161:	"5.01 Not Implemented",
            162:	"5.02 Bad Gateway",
            163:	"5.03 Service Unavailable",
            164:	"5.04 Gateway Timeout",
            165:	"5.05 Proxying Not Supported",

            # draft-ietf-core-block-14
            95:	"2.31 Continue",
            136:	"4.08 Request Entity Incomplete",
        }
    __responses_groups = (
        'Reserved',
        'Reserved',
        'Success',
        'Reserved',
        'Client Error',
        'Server Error',
    )


    def __getitem__ (self, item):
        try:
            return self.__known_codes[item]
        except KeyError:
            if 1 <= item <= 31:
                return "Request %d" % item
            elif 64 <= item <= 191:
                major = item // 32
                minor = item %  32
                return "%d.%02d %s" % (major, minor, self.__responses_groups[major])
            else:
                return "Reserved"

class CoAPType (UInt2):
    __values = "CON", "NON", "ACK", "RST"

    def __new__ (cls, value):
        if isinstance (value, str):
            try:
                value = cls.__values.index (value.upper())
            except ValueError:
                raise Exception ("Invalid CoAP message type")

        return super().__new__(cls, value)

class CoAPCode (UInt8):
    __values = "Empty", "GET", "POST", "PUT", "DELETE"

    def __new__ (cls, value):
        if isinstance (value, str):
            try:
                value = cls.__values.index (value.upper())
            except ValueError:
                raise Exception ("Invalid CoAP code")
        elif isinstance (value, float):
            major = int (value)
            minor = round ((value-major)*100)
            if not (0 <= major < 8) or not (0 <= minor < 32):
                raise Exception ("Invalid CoAP code")
            value = major*32 + minor

        return super().__new__(cls, value)

class _CoAPPayloadTag (PacketValue.Tag):
    def build_message (self, value, ctx = None):
        v, b = value.build_message()
        if b:
            # prepend the payload marker
            b = concatenate ((b"\xff", b))
        return v, b

    def decode_message (self, type_, bin_slice, ctx = None):
        if bin_slice and bin_slice[0] == 0xff:
            # payload present
            # -> skip the marker
            return type_.decode_message (bin_slice[1:])
        else:
            return b"", bin_slice


"""
     0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Ver| T |  TKL  |      Code     |          Message ID           |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |   Token (if any, TKL bytes) ...
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |   Options (if any) ...
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |1 1 1 1 1 1 1 1|    Payload (if any) ...
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
class CoAP (
    metaclass = InetPacketClass,
    fields    = [
        ("Version", 		"ver",		UInt2,		1),
        ("Type",		"type",		CoAPType,	0),
        ("TokenLength",		"tkl",		UInt4,		InetLength("Token")),
        ("Code", 		"code", 	CoAPCode,	0),
        ("MessageID", 		"mid",	 	Hex (UInt16),	0),
        ("Token",		"tok",		bytes,		b""),
        ("Options", 		"opt",	 	CoAPOptionList,	()),
        ("Payload", 		"pl",	 	Value,		_CoAPPayloadTag (b"")),
    ],
    descriptions = {
        "Version": {
            1:	"draft-ietf-core-coap",
        },
        "Type": {
            0:	"CON",
            1:	"NON",
            2:	"ACK",
            3:	"RST",
        },
        "Code": _CoAPCodeDescription(),
    }):

    @typecheck
    def is_request (self: is_flat_value):
        return 0 < self["code"] < 32

    @typecheck
    def is_response (self: is_flat_value):
        return self["code"] >= 32

    def get_uri (self, location = False):
        assert is_flat_value (self)
        host = ""
        port = ""
        path = ""
        query= ""
        for opt in self["opt"]:
            if not location:
                if isinstance (opt, CoAPOptionUriHost):
                    host = opt["val"]
                elif isinstance (opt, CoAPOptionUriPort):
                    port = str (opt["val"])
                elif isinstance (opt, CoAPOptionUriPath):
                    path += "/" + quote_plus (opt["val"])
                elif isinstance (opt, CoAPOptionUriQuery):
                    query += ("&" if query else "?") + quote_plus (opt["val"], safe = "=")
            else:
                if isinstance (opt, CoAPOptionLocationPath):
                    path += "/" + quote_plus (opt["val"])
                elif isinstance (opt, CoAPOptionLocationQuery):
                    query += ("&" if query else "?") + quote_plus (opt["val"], safe = "=")

        if not path and not location:
            path = "/"

        return "".join ((host, (":" if host or port else ""), port, path, query))

    def describe (self, desc):
        t = self.get_description ("type")
        if self["mid"]:
            t += " %d" % self["mid"]

        desc.info = "CoAP [%s] %s %s" % (
            t,
            self.get_description ("code"),
            self.get_uri (self["code"] >= 32) if self["code"] else "",
        )
        return True

#############################
# CoAP Options
#############################

##

class CoAPOptionUInt (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOption,
    prune		= -1,
    fields = [
        ("Value",	"val",	_CoAPUInt),
    ]):
    pass

class CoAPOptionEmpty (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOption,
    prune		= -1,
    ):
    pass

class CoAPOptionString (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOption,
    prune		= -1,
    fields = [
        ("Value",	"val",	str),
    ]):
    pass

class CoAPOptionBlock (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOption,
    prune		= -1,
    fields = [
        ("Number",		"num",	int,		0),
        ("M",			"m",	bool,		False),
        ("SizeExponent",	"szx",	UInt3,		0),
    ],
    descriptions	= {
        "szx": lambda szx: ("%d bytes" % (2**(szx+4))),
        "m": {
            0:	"last block",
            1:	"more blocks",
        },
    }):

    def __init__ (self, *k, **kw):
        # bypass CoAPOption.__init__
        return super (CoAPOption, self).__init__ (*k, **kw)

    def get_description_for_value (self, field, value):
        field = self.get_field_id (field)
        if field == 2: # 'Number' option
            szx = self["szx"]
            if value is None or szx is None:
                return None
            else:
                return "offset %d bytes" % (value * (2**(szx+4)))
        else:
            return super().get_description_for_value (field, value)

    @typecheck
    def _build_message (self) -> is_flatvalue_binary:

        values = self._fill_default_values()

        v = _CoAPUInt(
            # block number
            ((values[2] & 0xfffff) << 4)

            # m bit
            | ((values[3] & 1) << 3)

            # size exponent
            | (values[4] & 7)
        )

        v_bin = v.build_message()[1]

        if self["len"] is None:
            values[1] = len(v_bin)

        # encode the delta & length
        dl_bin = self._encode_delta_length(values[0], values[1])

        return type(self) (*values), (dl_bin + v_bin)

    @classmethod
    def _decode_body (cls, values, bin_slice):
        v, bin_slice = _CoAPUInt.decode_message(bin_slice)

        values.append ((v & ~0xf) >> 4)	# num
        values.append ((v & 8) >> 3)	# m
        values.append (v & 7)		# szx

        return bin_slice

class CoAPOptionEnd (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOptionEmpty,
    ):
    pass

##

_content_format_description = {
    0:	"text/plain; charset=utf-8",
    40:	"application/link-format",
    41:	"application/xml",
    42:	"application/octet-stream",
    47:	"application/exi",
    50:	"application/json",
    11542 : "application/vnd.oma.lwm2m+tlv",
    11543 : "application/vnd.oma.lwm2m+json",
}

_ResponseStutusCode_format_description = {
    65: "2001",
    66: "2002",
    67:	"2003",
    68:	"2004",
    69:	"2000",
}

_TY_format_description = {
    "AE": "2",
    "container": "3",
    "contentInstance": "4",
    "accessControlPolicy": "1",
    "subscription": "23",
    "node": "14",
    "remoteCSE": "16",
    "group": "9",
    "pollingChannel": "15",
    "flexContainer" : "28",
    "mgmtCmd": "12",   
}
def _max_age_description (v):

    if v == 0:
        return "no caching"

    result = []
    for div, name in (
        (60, "second"),
        (60, "minute"),
        (24, "hour"),
        (365, "day"),
        (99999999, "year"),
    ):
        nb = v %  div
        v  = v // div
        if nb:
            result.append ("%d %s%s" % (nb, name, "s" if nb>1 else ""))

    return ", ".join (reversed (result))

@classmethod
def _coap_option_decode_message (cls, bin_slice):

    # call the parent decoder
    v, bin_slice = cls.__bases__[0]._decode_message (bin_slice)
    v.__class__  = cls  # bless the object (to match our type)

    # ensure that the length of the option is valid
    lmin, lmax = cls._min_max_length
    if not (lmin <= v["len"] <= lmax):
        raise Exception ("Option %s has invalid length %d (should be in [%d..%d])" % (cls.__name__, v["len"], lmin, lmax))

    return v, bin_slice

for i, n, t, l, d in (
        # draft-ietf-core-coap-18
        # (MaxAge is defined separately)

        #No	Name			ParentClass	Min/Max length	description
        (1,	"IfMatch",		"",		(0, 8),		None),
        (3,	"UriHost",		"String",	(1, 255),	None),
        (4,	"ETag",			"",		(1, 8),		None),
        (5,	"IfNoneMatch",		"Empty",	(0, 0),		None),
        (7,	"UriPort",		"UInt",		(0, 2),		None),
        (8,	"LocationPath",		"String",	(0, 255),	None),
        (11,	"UriPath",		"String",	(0, 255),	None),
        (12,	"ContentFormat",	"UInt",		(0, 2),		_content_format_description),
        (14,	"MaxAge",		"UInt",		(0, 4),		_max_age_description),
        (15,	"UriQuery",		"String",	(0, 255),	None),
        (17,	"Accept",		"UInt",		(0, 2),		_content_format_description),
        (20,	"LocationQuery",	"String",	(0, 255),	None),
        (35,	"ProxyUri",		"String",	(1, 1034),	None),
        (39,	"ProxyScheme",		"String",	(1, 255),	None),
        (60,	"Size1",		"UInt",		(0, 4),		None),

        # draft-ietf-core-block-14

        (27,	"Block1",		"Block",	(0, 3),		None),
        (23,	"Block2",		"Block",	(0, 3),		None),
        (28,	"Size2",		"UInt",		(0, 4),		None),

        # draft-ietf-core-observe-11

        (6,	"Observe",		"UInt",		(0, 3),		None),

        # OneM2M binding
        (256, "OneM2MFrom" , "String" , (0, 255) , None),
        (257, "OneM2MRequestIdentifier" , "String" , (0, 255) , None),
	(258, "OneM2MName" , "String" , (0, 255) , None),
        (263, "OneM2MNotificationURI" , "String" , (0, 255) , None),
        (265, "OneM2MResponseStatusCode" , "UInt" , (0, 2) , _ResponseStutusCode_format_description),
        (266, "OneM2MGroupRequestIdentifier" , "String" , (0, 255) , None),
        (267, "OneM2MTY" , "UInt" , (0, 2) , _TY_format_description),
    ):
    exec(
"""
class CoAPOption%s (
    metaclass	= InetPacketClass,
    variant_of	= CoAPOption%s,
    id		= %d,
    descriptions	= {} if d is None else {"Value": d},
    ):
    _min_max_length = %r
    _decode_message = _coap_option_decode_message
""" % (n, t, i, l))


##Aliases
#ICMPv6NSol = ICMPv6NeighborSolicitation
#ICMPv6NAdv = ICMPv6NeighborAdvertisement
#ICMPv6RSol = ICMPv6RouterSolicitation
#ICMPv6RAdv = ICMPv6RouterAdvertisement
#ICMPv6EReq = ICMPv6EchoRequest
#ICMPv6ERep = ICMPv6EchoReply
#ICMPv6Unre = ICMPv6DestinationUnreacheable
#ICMPv6SLL = ICMPv6SLLOption
#ICMPv6TLL = ICMPv6TLLOption
#ICMPv6PI = ICMPv6PIOption


# tell the udp module on which udp ports coap runs
for port in (5683,) + tuple (range (61616, 61632)):
    ttproto.core.lib.inet.udp.udp_port_map[port] = CoAP




#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numbers, threading, time, types
from	ttproto.core.typecheck	import *

from ttproto.core import exceptions

# TODO: split into several modules
_all__ = [
	'Clock',
	'EventQueueClock',
	'SystemClock',
	'SimulatedClock',
]

class Clock:
	# TODO: maybe support multiple clocks in the same process ?
	__instance = None
	__lock = threading.Lock()

	@staticmethod
	def get_instance():
		return Clock.__instance

	@staticmethod
	@typecheck
	def set_instance (clock: this_class):
		with Clock.__lock:
			if Clock.__instance:
				# FIXME: what to do with the scheduled events ?
				#	 maybe it would be better to raise an exception here
				Clock.__instance.kill()

			Clock.__instance = clock

	def unschedule_events (self, callback: types.MethodType):
		raise exceptions.NotImplemented()

	@typecheck
	def schedule_event_absolute (self, time: numbers.Number, callback: callable):
		raise exceptions.NotImplemented()

	@typecheck
	def schedule_event_relative (self, time: numbers.Number, callback: callable):
		raise exceptions.NotImplemented()

	def reset (self):
		raise exceptions.NotImplemented()

	def kill (self):
		pass

	def time (self):
		raise exceptions.NotImplemented()

	def all_threads_blocked (self):
		"""Called by SnapshotManager when all threads are waiting for an external event"""
		pass

def _Condition (lock):
	"""Wrapper for threading._Condition() implementing the result value in
	Condition.wait() (as implemented in Python 3.2)

	(to be removed once we upgrade to python 3.2)

	NOTE: this wrapper won't work correctly if multiple threads are calling
	wait() concurrently

	NOTE: this wrapper won't work with notify_all()
	"""

	cond = threading.Condition (lock)

	super_notify	= cond.notify
	super_wait	= cond.wait

	notify_received	= False

	def notify():
		nonlocal notify_received

		notify_received = True
		super_notify()

	def wait (timeout = None):
		nonlocal notify_received

		notify_received = False
		super_wait (timeout)
		return notify_received

	cond.notify	= notify
	cond.wait	= wait

	return cond


class EventQueueClock (Clock):
	"""An event is stored as a tuple (timestamp, time) and are ordered chronologically in the queue"""

	def __init__ (self):
		Clock.__init__ (self)

		self.__queue = []
		self._lock = threading.Lock()
		self._cond = _Condition(self._lock)

	@typecheck
	def schedule_event_relative (self, time: numbers.Number, callback: callable):
		with self._lock:
			self.__new_event (self._time() + time, callback)

	@typecheck
	def schedule_event_absolute (self, time: numbers.Number, callback: callable):
		with self._lock:
			self.__new_event (time, callback)

	def unschedule_events (self, callback):
		with self._lock:
			q = self.__queue
			i = len (q) - 1
			while i>=0:
				if q[i][1] == callback:
					q.pop(i)
				i-=1

			self._cond.notify()

	def reset (self):
		with self._lock:
			self.__queue[:] = ()

			self._cond.notify()

	def __new_event (self, time, callback):
		assert self._lock.locked()

		diff = self._time() - time
		if diff > 0:
			print ("WARNING: setting an event in the past (%e seconds in the past)" % diff)
			#TODO: allow a jitter margin, raise an exception if over the threshold

		for i in range (0, len (self.__queue)):
			if time < self.__queue[i][0]:
				self.__queue.insert (i, (time, callback))
				break
		else:
			self.__queue.append ((time, callback))

		self._cond.notify()

	@typecheck
	def _get_first_event (self) -> optional (tuple):
		assert self._lock.locked()

		return self.__queue[0] if len (self.__queue) else None

	def _shift (self):
		assert self._lock.locked()
		assert len (self.__queue)

		self.__queue.pop(0)

	def time (self):
		with self._lock:
			return self._time()

	def _time (self):
		assert self._lock.locked()

		raise exceptions.NotImplemented()


class SystemClock (EventQueueClock):
	def __init__ (self):
		EventQueueClock.__init__ (self)
		self.__thread = threading.Thread (target = self.__thread_func)
		self.__thread.daemon = True
		self.__thread.start()

	def kill (self):
		with self._lock:
			assert self.__thread

			thread = self.__thread
			self.__thread = None

			self._cond.notify()

		thread.join()

	def _time (self):
		assert self._lock.locked()

		return time.time()


	def __thread_func (self):

		while True:
			with self._lock:
				if self.__thread != threading.current_thread():
					return

				ev = self._get_first_event()

				if ev:
					remaining_time = ev[0] - self._time()
					if remaining_time > 0:
						self._cond.wait (remaining_time)
						continue
				else:
					self._cond.wait()
					continue

				self._shift()

			# report the event
			ev[1]()

class SimulatedClock (EventQueueClock):
	def __init__ (self):
		EventQueueClock.__init__ (self)

		self.__current_time = None

		self.__killed = False

		self.__thread = threading.Thread (target = self.__thread_func)
		self.__thread.daemon = True
		self.__thread.start()

	def kill (self):
		with self._lock:
			self.__killed = True
			self._cond.notify()

		self.__thread.join()

	def _time (self):
		assert self._lock.locked()

		if self.__current_time is None:
			return 0 # FIXME: should have a better way to do it
			raise exceptions.Error ("Simulated clock not yet started")
		else:
			return self.__current_time

	def reset (self):
		with self._lock:
			self.__current_time = None
		super().reset()

	def start (self):
		with self._lock:
			ev = self._get_first_event()

			self.__current_time = ev[0] if ev else 0

	def __thread_func (self):
		from ttproto.core import snapshot

		with self._lock:
			while True:
				self._cond.wait()

				if self.__killed:
					return

				if self.__current_time is None:
					# clock not yet started
					# -> wait for the signal
					continue

				# clock is running

				while True:
					ev = self._get_first_event()
					if not ev:
						# no events in the queue
						if snapshot.SnapshotManager.all_tracked_threads_blocked():
							# TODO: stop the execution
							print ("Error: deadlock detected")
						break

					if (ev[0] <= self.__current_time):
						# the event is due
						# -> remove it from the list and report it
						self._shift()

						try:
							self._lock.release()
							ev[1]()
						finally:
							self._lock.acquire()

						if self.__current_time is None:
							break

					elif snapshot.SnapshotManager.all_tracked_threads_blocked():
						# all threads are in waiting state
						if ev:
							if ev[0] > self.__current_time:
								# -> advance the clock to the time of the next event
								self.__current_time = ev[0]
					else:
						# other threads are running
						# -> wait
						break




	def all_threads_blocked (self):
		with self._lock:
			self._cond.notify()


Clock.set_instance (SystemClock())


#raise "Je pense que j'ai un gros deadlock entre SimulatedClock et SnapshotManager"



from ..common import *


class TD_M2M_NH_51(CoAPTestCase):
    """
---
TD_M2M_NH_51:
    cfg: M2M_CFG_01
    obj: AE sends a <contentInstance> CREATE request to a <container> which contains attribute currentNrOfInstances whose value equals to that of maxNrOfInstances and Registrar CSE deletes
         the oldest <contentInstance> from the parent <container> and then creates the requested <contentInstance> resource for the originator AE.
    pre: AE has created an application resource <AE> on registrar CSE and AE has created a <container> resource on registrar CSE (where the number of contentInstances equals to the value
         set in maxNrOfInstance) on registrar CSE.
    ref: 'TS-0001, clause 10.2.19.2 ; TS-0004, clause 7.3.6.2.1'
    seq:
    -   s: "AE sends a RETRIEVE request with resultContent set to 1 (default value) to retrieve the <oldest> contentInstance resource and AE sends a request to CREATE a <contentInstance> resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON
            - Code = 2 (POST)
            - Non-empty payload
            - Content-format option
            - oneM2M-TY "4"
            - oneM2M-FR (originator) "AE_ID"
            - oneM2M-RQI "token-string" (CRQI)
    -   v: Check if possible that the <oldest> resource of <container> resource is deleted. 
    -   c:
        - 'Registrar CSE sends response containing:'
        -   - Code = 2.01 (Created)
            - oneM2M-RSC= 2001
            - oneM2M-RQI=CRQI
            - Location-Path
            - contentFormat option
            - non-empty payload
    
    -   v: AE indicates successful CREATE operation of <contentInstance> and indicates the representation of the recent <oldest> resource in the <container> is different with that of <oldest>
           resource retrieved at the beginning of test in terms of resourceID and resourceName attribute value.
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [CoAP(type='con', code='post')]


    def run (self):

        
        self.match('client', CoAP (type="con", code="post",pl=Not(b"")), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MFrom())), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MTY('4'))), 'fail')
        if self.match('client', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier())), 'fail'): 


            CMID = self.coap['mid']
            CTOK = self.coap['tok']
            OPTS = self.coap['opt']
            RI = OPTS[CoAPOptionOneM2MRequestIdentifier]
            RIVAL = RI[2]

            self.next()

            self.match('server', CoAP(code=2.01, mid=CMID, tok=CTOK, pl=Not(b'')), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MResponseStatusCode('2001'))), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionOneM2MRequestIdentifier(RIVAL))), 'fail')
            self.match('server', CoAP(opt=Opt(CoAPOptionLocationPath())), 'fail')



#!/usr/bin/env python3

from ..common import *


class TD_COAP_CORE_18 (CoAPTestCase):
    """
---
TD_COAP_CORE_18:
    cfg: CoAP_CFG_BASIC
    obj: Perform POST transaction with responses containing several
         Location-Path options (CON mode)
    pre: Server accepts creation of new resource on /test and the created
        resource is located at /location1/location2/location3 (resource
        does not exist yet)
    ref: '[COAP] 5.8.1, 5.10.8, 5.9.1.1'
    seq:
    -   s: "Client is requested to send a confirmable POST request to server\
            \u2019s resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 0 (CON
            - Code = 2 (POST)
            - An arbitrary payload
            - Content-format option
            - Uri-Path option "test"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.01 (Created)
        - 'and three options of type Location-Path, with the values (none
            of which contains a "/"):'
        -   - location1
            - location2
            - location3
    -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [
            CoAP(type='con', code='post')
        ]

    request_uri = "/test"

    def run (self):
        self.match ("client", CoAP (type="con", code = "post",pl=Not(b""),
                    opt=self.uri(
                        self.request_uri,
                        CoAPOptionContentFormat(),
                    )))

        self.next_skip_ack()
        #TODO: generate a fail if we have a '/' or '.' or '..' in the CoAPOptionLocationPath ? see 5.10.7
        self.match ("server", CoAP (code=2.01,
                        opt=Opt(
                            CoAPOptionLocationPath ("location1"),
                            CoAPOptionLocationPath ("location2"),
                            CoAPOptionLocationPath ("location3"),
                )))

        self.next_skip_ack (optional = True)




#!/usr/bin/env python3

#from ttproto.ts_coap.proto_specific import CoAPTestcase
# from ttproto.ts_coap.proto_templates import *
from ttproto.ts_coap.testcases.td_coap_core_23 import TD_COAP_CORE_23

class TD_COAP_CORE_28 (TD_COAP_CORE_23):
	"""Identifier:
TD_COAP_CORE_28
Objective:
Perform GET transaction with responses containing the ETag option and requests
containing the If-None-Match option (CON mode) (Reverse proxy)
Configuration:
CoAP_CFG_03
References:
[1] clause 5.8.1, 5.10.7,5.10.10,12.1.12, 8.2.2,8.2.1,10.2.2,11.2

Pre-test
conditions:
•	 Proxy is configured as a reverse-proxy for the server
•	Proxy’s cache is cleared
•	Server should offer a /create3 resource, which does not exist and which can be
created by the client
•	Client & server supports If-None-Match


Test Sequence:
Step
Type
Description
Part A: single creation


1
Stimulus
Client is requested to send a confirmable PUT request to
proxy to atomically create resource in server

2
Check
Proxy forwards the request to server


3
Check
Forwarded t request must contain:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Option Type=If-None-Match
•	An arbitrary payload

4
Check
Server sends response containing:
•	Code = 65 (2.01 Created)


5
Check
Proxy forwards the response to client

6
Verify
Client displays the response & and server created new
resource
Part B: concurrent creations


5
Stimulus
Client is requested to send s confirmable PUT request to
proxy to atomically create resource in server

6
Check
Sent request must contain:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Option Type=If-Non-Match
•	Option value=Received ETag value

7
Check
Server sends response containing:
•	140 (4.12 Precondition Failed)

8
Verify
Proxy forwards the response to client

9
Check
Response contains:
•  140 (4.12 Precondition Failed)

10
Verify
Client displays the response
	"""
	request_uri = "/create3"
	reverse_proxy = True



import os
import time
import json
import base64
import hashlib
import logging

from collections import OrderedDict

from ttproto import LOG_LEVEL
from ttproto.core.lib.all import *
from ttproto.core.analyzer import Analyzer
from ttproto.core.dissector import Capture, get_dissectable_protocols
from ttproto.core.typecheck import typecheck, optional, either
from ttproto.utils.pcap_filter import remove_first_frames

ALLOWED_PROTOCOLS_FOR_ANALYSIS = ['coap', '6lowpan', 'onem2m', 'lwm2m']

# Directories
DATADIR = "data"
TMPDIR = "tmp"
LOGDIR = "log"

# Prefix and suffix for the hashes
HASH_PREFIX = 'tt'
HASH_SUFFIX = 'proto'
TOKEN_LENGTH = 28

# # # AUXILIARY FUNCTIONS # # #

logger = logging.getLogger('tat|ttproto_api')
logger.setLevel(LOG_LEVEL)


def analyze_capture(filename, protocol, testcase_id, output_file):
    """
    Analyses network traces (.pcap file) based on the test cases checks.
    """
    assert filename
    assert protocol
    assert testcase_id

    if os.path.isfile(filename) is False and os.path.isfile(os.path.join(TMPDIR, filename)):
        filename = os.path.join(TMPDIR, filename)

    logger.info("Analyzing PCAP file %s" % filename)

    if protocol.lower() not in ALLOWED_PROTOCOLS_FOR_ANALYSIS:
        raise NotImplementedError('Protocol %s not among the allowed analysis test suites' % protocol)

    analysis_results = Analyzer('tat_' + protocol.lower()).analyse(filename, testcase_id)
    logger.info('PCAP file analysed: %s' % filename)
    logger.info('Analysis result: %s' % str(analysis_results))

    if output_file and type(output_file) is str:
        # save analysis response
        _dump_json_to_file(json.dumps(analysis_results), output_file)
        logger.info('Results saved at: %s' % output_file)

    return analysis_results


def dissect_capture(filename, proto_filter=None, output_file=None, number_of_frames_to_skip=None):
    """
    Dissects (decodes and converts to string representation) network traces (.pcap file).
    """
    assert filename

    if os.path.isfile(filename) is False and os.path.isfile(os.path.join(TMPDIR, filename)):
        filename = os.path.join(TMPDIR, filename)

    logger.info("PCAP file dissection starts. Filename: %s" % filename)

    proto_matched = None

    if proto_filter:
        # In function of the protocol asked
        proto_matched = get_protocol(proto_filter)
        if proto_matched is None:
            raise Exception('Unknown protocol %s' % proto_filter)

    if number_of_frames_to_skip:
        filename_pcap_filtered = os.path.join(TMPDIR, 'pcap_without_some_skipped_frames.pcap')
        remove_first_frames(
            pcap_filename=filename,
            new_pcap_filename=filename_pcap_filtered,
            number_of_frames_to_skip=number_of_frames_to_skip
        )
        filename = filename_pcap_filtered

    cap = Capture(filename)

    if proto_matched and len(proto_matched) == 1:
        print(proto_matched)
        proto = eval(proto_matched[0]['name'])
        dissection_as_dicts = cap.get_dissection(proto)
        dissection_as_text = cap.get_dissection_simple_format(proto)
        frames_summary = cap.frames
    else:
        dissection_as_dicts = cap.get_dissection()
        dissection_as_text = cap.get_dissection_simple_format()
        frames_summary = cap.frames

    if frames_summary:
        logger.info('PCAP file dissected (filename: %s). Frames summary:\n%s' % (
            filename,
            json.dumps(([repr(c) for c in frames_summary]), indent=4)
        ))
    else:
        logger.info('PCAP file dissected (filename: %s). No frames found.')

    if output_file and type(output_file) is str:
        # save dissection response
        _dump_json_to_file(json.dumps(dissection_as_dicts), output_file)

    return dissection_as_dicts, dissection_as_text


@typecheck
def get_test_cases(
        testcase_id: optional(str) = None,
        verbose: bool = False
) -> OrderedDict:
    """
    Function to get the implemented test cases

    :param testcase_id: The id of the single test case if one wanted
    :param verbose: True if we want a verbose response
    :type testcase_id: str
    :type verbose: bool

    :return: The implemented test cases using doc.f-interop.eu format
    :rtype: OrderedDict
    """

    test_cases = OrderedDict()
    tc_query = [] if not testcase_id else [testcase_id]
    raw_tcs = Analyzer('tat_coap').get_implemented_testcases(
        tc_query,
        verbose
    )

    # Build the clean results list
    for raw_tc in raw_tcs:
        tc_basic = OrderedDict()
        tc_basic['_type'] = 'tc_basic'
        tc_basic['id'] = raw_tc[0]
        tc_basic['objective'] = raw_tc[1]

        tc_implementation = OrderedDict()
        tc_implementation['_type'] = 'tc_implementation'
        tc_implementation['implementation'] = raw_tc[2]

        # Tuple, basic + implementation
        test_cases[raw_tc[0]] = OrderedDict()
        test_cases[raw_tc[0]]['tc_basic'] = tc_basic
        test_cases[raw_tc[0]]['tc_implementation'] = tc_implementation

    # If a single element is asked
    if testcase_id:
        test_cases = test_cases[raw_tcs[0][0]]

    # Return the results
    return test_cases


@typecheck
def get_protocol(
        protocol: optional(str) = None
) -> either(list, type(None)):
    """
    Function to get the protocols protocol(s) info dict

    :param protocol: The name of the protocol
    :type protocol: str

    :return: list of implemented protocols, conditioned to the protocol filter
    :rtype: list of OrderedDict(s)
    """

    # list to return
    answer = []

    # Getter of protocol's classes from dissector
    prot_classes = get_dissectable_protocols()
    logger.debug(str(prot_classes))

    # Build the clean results list
    for prot_class in prot_classes:

        if protocol and protocol.lower() == prot_class.__name__.lower():
            # Prepare the dict for the answer
            prot = OrderedDict()
            prot['_type'] = 'implemented_protocol'
            prot['name'] = prot_class.__name__
            prot['description'] = ''
            return [prot]
        elif protocol is None:
            # Prepare the subdir for the dict-inception (answer)
            prot = OrderedDict()
            prot['_type'] = 'implemented_protocol'
            prot['name'] = prot_class.__name__
            prot['description'] = ''
            answer.append(prot)
        else:
            # not the selected one
            pass

    if answer is None or len(answer) == 0:
        return None
    else:
        return answer


def base64_to_pcap_file(filename, pcap_file_base64):
    """
    Returns number of bytes saved.

    :param filename:
    :param pcap_file_base64:
    :return:
    """
    # save to file
    with open(filename, "wb") as pcap_file:
        nb = pcap_file.write(base64.b64decode(pcap_file_base64))

        return nb


@typecheck
def get_protocols_list() -> list:
    """Function to get the protocols protocols list"""
    return [p.__name__ for p in get_dissectable_protocols()]


def _dump_json_to_file(json_object, filename):
    """

    :param json_object:
    :param filename: filename must include PATH
    :return:
    """

    if '.json' not in filename:
        filename += '.json'

    with open(filename, 'w') as f:
        f.write(json_object)


@typecheck
def _get_token(tok: optional(str) = None):
    """
    Function to get a token, if there's a valid one entered just return it
    otherwise generate a new one

    :param tok: The token if there's already one
    :type tok: str

    :return: A token, the same if there's already one, a new one otherwise
    :rtype: str
    """

    # If the token is already a correct one
    try:
        if all((
                tok,
                type(tok) == str,
                len(tok) == 28,
                base64.urlsafe_b64decode(tok + '=')  # Add '=' only for checking
        )):
            return tok
    except:  # If the decode throw an error => Wrong base64
        pass

    # Generate a token
    token = hashlib.sha1(
        str.encode((
            "%s%04d%s" %
            (
                HASH_PREFIX,
                time.time(),
                HASH_SUFFIX
            )
        ), encoding='utf-8')
    )
    token = base64.urlsafe_b64encode(token.digest()).decode()

    # Remove the '=' at the end of it, it is used by base64 for padding
    return token.replace('=', '')


if __name__ == "__main__":
    proto_class = get_protocol('CoAP')
    # print(str(proto_class))

    proto_class = get_protocol()
    # print(str(proto_class))

    testcases = get_test_cases()
    # print(str(testcases))

    dissection = dissect_capture(
        filename='tests/test_dumps/dissection/coap/CoAP_plus_random_UDP_messages.pcap'
    )
    # print(dissection)

    p_list = get_protocols_list()
    print(p_list)



#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
import numbers, threading, traceback
from .common import *

from ttproto.core import	clock, snapshot, exceptions

from ttproto.core.templates import Superset #FIXME: why we need it ?!

__all__ = [
	"SixLowpanImplementation",
]

class Scheduler (snapshot.EventSource):
	class Event:
		@typecheck
		def __init__ (self, process, name: optional(str)):
			self.process = process
			self.name = name


	def __init__ (self, clk: optional (clock.Clock) = None):
		super().__init__()

		self.__clock = clk if clk is not None else clock.Clock.get_instance()
		self.event = snapshot.Event (self, "event")

		self.__events = []
		self.__expired_events = []
		
	
	def reset (self):
		with self.lock:
			for e in self.__events:
				self.__clock.unschedule_events (e)
			self.__events  = []
			self.__expired_events = []
	
	@typecheck
	def __find_event (self, process, name: str):
		for e in self.__events:
			if e.process == process and e.name == name:
				return e
	
	@typecheck
	def __unschedule_event (self, process, name: str):
		ev = self.__find_event (process, name)
		if ev is not None:
			self.__clock.unschedule_events (ev)
			self.__events.remove (ev)
			try:
				self.__expired_events.remove (ev)
			except ValueError:
				pass
	
	@typecheck
	def __prepare_schedule_event (self, process, name: optional(str) = None):
		if name is not None:
			self.__unschedule_event (process, name)

		ev = self.Event (process, name)
		self.__events.append (ev)
		def callback():
			with self.lock:
				self.__expired_events.append (ev)
			
		return callback
	
	@typecheck
	def schedule_event_relative (self, time: numbers.Number, process, name: optional(str) = None):
		with self.lock:
			self.__clock.schedule_event_relative (time, self.__prepare_schedule_event (process, name))

	@typecheck
	def schedule_event_absolute (self, time: numbers.Number, process, name: optional(str) = None):
		with self.lock:
			self.__clock.schedule_event_absolute (time, self.__prepare_schedule_event (process, name))
	
	def evaluate_snapshot(self):
		self.__snapshot_expired_events = len (self.__expired_events)

	def time (self):
		return self.__clock.time()
	
	class __EventMatch (snapshot.EventMatch):
		def __init__ (self, snapshot_event, ev):
			super().__init__ (snapshot_event)
			self.__event = ev

		def get_ev (self):
			return self.__event

	def match_event (self):

		while self.__snapshot_expired_events:

			self.__snapshot_expired_events -= 1
			ev = self.__expired_events.pop(0)
			try:
				self.__events.remove (ev)
			except ValueError:
				#print ("WARNING: FIXME: unscheduled event still present")
				pass
			else:
				return self.__EventMatch (self.event, ev)
	
		return None


class SixLowpanImplementation (snapshot.SnapshotManager.Thread):

	class ConfigClass:
		pass
	config = ConfigClass()
	config.debug = 1
	config.eui64 = "0200000001020304"
	config.routeraddr = ""

	#delay we use when we schedule a RS
	config.DelayRS = 0.1

	#delay we use when we schedule a NS
	config.DelayNS = 0.1

	# time to refresh before a router expire
	config.routerRefreshTime = 6
	# time to refresh before an address expire
	config.AddressRefreshTime = 6


	#FIXME # list of prefix
	PRFX1 = "2001:0db8:1:1::"
	PRFX2 = "fdea:1ffc:f851::"
	list_of_prefix = [PRFX1,PRFX2]


	class Process:
	
		def __init__ (self, implem, func, *k, **kw):
			"""Initialise the new process
			
			- implem	the implementation (python object) hosting this process 
			- function	generator function implementing the behaviour of this process

			"""
			self.__generator  = func (self, *k, **kw)
			self.__scheduler  = implem.get_scheduler()
			next (self.__generator)

		def send (self, msg):
			"""Report an event (message) to the process"""

			return self.__generator.send (msg)

		@typecheck
		def schedule_event_relative (self, time: numbers.Number, name: optional(str) = None):
			"""Schedule an event in the future.

			- time		the time at which the event is to be reported
			- name		name of the event (optional)

			The time is given in seconds starting from now.

			The name is optional. If there is already an existing
			event with the same name for the same process, then
			this existing event is overwritten by the new one.		
			"""
			self.__scheduler.schedule_event_relative (time, self, name)

		@typecheck
		def schedule_event_absolute (self, time: numbers.Number, name: optional(str) = None):
			"""Schedule an event in the future.

			- time		the time at which the event is to be reported
			- name		name of the event (optional)

			The time is given in seconds starting from the epoch (1st Jan 1970).
ig.delayRS, 
			The name is optional. If there is already an existing
			event with the same name for the same process, then
			this existing event is overwritten by the new one.		
			"""
			self.__scheduler.schedule_event_absolute (time, self, name)

	@typecheck
	def __init__ (self, eui64 = "0200000102030405", clock: optional (clock.Clock) = None):
		super().__init__()
		self.__eui64 = store_data (eui64, Eui64Address)
		self.__port  = SixLowpanMessagePort (0)

		self.__scheduler = Scheduler (clock)

	# FIXME: function needed to avoid log errors
	def log_event (self, event):
		pass

	def get_scheduler (self):
		return self.__scheduler

	def get_message_port (self):
		return self.__port
	
	def reset (self):
		# clear all scheduled events
		self.__scheduler.reset()

		# generate the link local address (from the eui-64 id)
		self.__lladdr = IPv6Prefix("fe80::").make_address(self.__eui64)

		# by default we listen to two addresses:
		#  - our link-local address
		#  - the all-nodes multicast group ff02::2
		#
		# NOTE: the addresses are stored in a regular list (instead of
		# 	a set). Thus it is possible to have duplicates (which
		# 	is useful because we may register the same address with
		# 	multiple routers but we should not remove completely
		# 	the address when one router gets unreachable)
		self.__addresses = [self.__lladdr,IPv6Address("ff02::2")]


		# clear the message handlers
		self.__message_handlers = [] 
		
		# launch the processes
		self.__rdisc  = self.Process (self, self.__router_discovery_process)
		self.__icmpv6 = self.Process (self, self.__icmpv6_process)

		# dictionnary containing the process spawned for each router
		#	key:	IPv6 link local address of the router
		#	value:	register_process
		self.__router_processes = {}

		# blacklist contain all duplicated address
		self._black_list_addrDupl = []

		# blacklist contain all routeur with a full cache
		self._black_list_routeurFull = []
		


	def debug(self, msg):
		if self.config.debug:
			print (msg)
	
	def register_message (self, pattern, process):
		"""
		
		"""
		self.__message_handlers.append ((pattern, process))
	
	def run (self):
		"""Main loop
		
		The main loops monitors external events:
		 - messages received on the communication port
		 - scheduled event
		
		Then it reports the event to the adequate process.
		 - scheduled events are reported to the process that created
		   this event (by calling Process.schedule_event_xxxxxxx())
		 - messages are compared with the patterns registered (using
		   the register_message() function) and in case of a match, the
		   message is reported to the corresponding process (in case of
		   multiple matches, the message is reported multiple times)

		Events are reported to the process using the send() method
		(thus they can be received using the yield instruction).
		"""
		
		self.reset()

		while True:
			with alt:

				# match a scheduled event
				@self.__scheduler.event
				def _(ev):
					ev.process.send (ev.name)		

				# receive an IPv6 message from the communication port
				@self.__port.receive (IPv6())
				def _(value):
					# we get the IPv6 datagram
					ipv6_msg = value["pl"]["pl"]

					# filter out packets not addressed to this node
					if ipv6_msg["dst"] not in self.__addresses:
						print ("Addresses:", self.__addresses)
						print ("Ignored message (not for us):", ipv6_msg)
						return

					# check if one process is expecting this message
					# and report it if any
					for pattern, process in tuple(self.__message_handlers):
						if pattern.match (ipv6_msg):
							# report the IPv6 message to the process
							# (will be the result of the yield instruction)
							process.send (ipv6_msg)

				# receive any other message
				@self.__port.receive
				def _(value):
					print ("6lowpan implem: unexpected message:", value)
			
	def __send (self, msg):
		"""Send a message to the communication port

		This function chooses the destination (hw address) according to
		the IPv6 destination address and return True if successful.

		It will return False if the port is not connected or if there
		is no adequate route in the routing table.

		Note: it is not recommended to manipulate directly self.__port
		(eg: by calling self.__port.send()) because it may raise an
		exception if the port is not connected. This function will
		silently catch the exceptions.
		"""

		msg = msg.flatten()

		try:
			assert isinstance (msg, IPv6) # TODO: support 6lowpan & IPHC ?

			if msg["dst"][0] in (0xfe, 0xff): # FIXME: global multicast not supported
				# link-local or multicast destination
				self.__port.send (msg)
			else:
				# global destination
				# -> use the first known router
				if not self.__router_processes:
					# no route!
					return False

				self.__port.send (msg, dst = next(iter(self.__router_processes.keys())))

			return True

		except exceptions.PortNotConnected:
			return False

	def __send_ra (self, ipv6_destination_address):
		self.__send (
			IPv6 (src=self.__lladdr, dst=ipv6_destination_address) /
			ICMPv6RAdv (rlt = 30 , hp = 255 , opt =[ICMPv6SLL (hw = self.__eui64),SixPI( pf=self.PRFX1 , l = 0 ) ,]))

	def __send_rs (self, ipv6_destination_address):
		self.__send (
			IPv6 (src=self.__lladdr, dst=ipv6_destination_address) /
			ICMPv6RSol (opt =[ICMPv6SLL (hw = self.__eui64)]))




##########################################################################################################################
###################################################### ROUTER DISCOVERY PROCESS ##############################################
#########################################################################################################################



	def __router_discovery_process (self, process):
		"""Router Discovery Process
		
		This process monitors the Router Advertisement received on the
		link so as to maintain the router list up-to-date. It spawns
		one register process for each router.

		If the router list is empty, then it will send and retransmit
		Router Solicitations.
		"""

		# Register the messages to be handled by this process (RS & RA)
		#FIXME can be must precised	
		rs_template = IPv6()/ICMPv6RSol()
		ra_template = IPv6()/ICMPv6RAdv()
		
		self.register_message (rs_template, process)
		self.register_message (ra_template, process)

		# dict containing the expiration time for each router
		router_expiry_times = {}

		# number of retransmissions done so far
		retransmission_count = 0

		
		while True:

			msg = yield
							
			#-------------------- We receive a Router Solicitation from an HOTE  ------------------#
				
			if (rs_template.match (msg)):
				# we send a router solicitation to the LBR
				self.__send_rs ()



				if ( (msg["src"] in self._black_list_addrDupl) or (msg["src"] in self.__router_processes.keys()) ) :
					print("Address already used")
				
				else :
					# spawn the new process
					p = self.Process (self, self.__router_registration_process, hote_addr)
					self.__router_processes[hote_addr] = p
			
			#-------------------- We receive a Router Advertisement from a LBR  ------------------#
			if (ra_template.match (msg)):
				# we send a router advertisement to the HOTE
				self.__send_ra (msg["dst"])
				
					



##########################################################################################################################
###################################################### REGISTRATION PROCESS ##############################################
##########################################################################################################################



	def __router_registration_process (self, process, hote_addr):
		
		
		

		# Register the messages to be handled by this process (NS)	
		self.register_message (IPv6()/ICMPv6NSol(), process)

		#
		self.eui64 = "00000000000000000"


		# expiration time of the router
		# (sys.maxsize if none)
		router_expiry_time = sys.maxsize

		# boolean used to ensure that we send at most one unicast RS in
		# case the router information need to be refreshed (FIXME: shall we retransmit it ?)
		unicast_rs_sent = True

		# list of prefix info classes
		#	key:	prefix
		#	value:	PrefixInfo
		prefix_infos = {}


		# list of context info classes
		#	key:	context
		#	value:	ContextInfo
		context_infos = {}

		# list of prefix infos for which we want to register an address
		register_queue = []
			

		# PrefixInfo being registered
		pending_registration = None





		def __send_na (self, hote_addr, prefix_info):

		
			self.__send (
				IPv6(src = self.__lladdr, dst = hote_addr ) /
				ICMPv6NAdv (opt = [
					ICMPv6SLL ( hw=self.eui64 ) ,
					SixARO( lt = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0 , 
					eui= ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)




		while True:
			
			print ("Reg: yield")
			msg = yield
			try:
				# TODO: wake-up to renew/revoke addresses
				now = self.__scheduler.time()
				
				print ("Reg:", now, msg)

				#-------------------- termination signal -------#

				if msg == "Kill":
					print ("Killing registration process for", hote_addr)
					
					# deconfigure the registered addresses
					for info in prefix_infos.values():
						if info.registered:
							self.__addresses.remove (info.address)

					while True:
						yield "Finished"
				
				#-------------------- it is time to refresh the router informations -------#


				#-------------------- When we receive a Neighbor Solicitation ------------------#

				elif (IPv6()/ICMPv6NSol()).match (msg):
	
					if msg[ICMPv6]["opt"][SixARO]["lt"] == 0 :
						# if the lifetime is 0, we must kill the process
						process.schedule_event_relative (-10,"Kill")
					else :
						self.eui64 = msg[ICMPv6]["opt"][SixARO]["eui"]	
						
						# we received a Neighbor Solicitation
						self.__send_na (msg["src"])


				#-------------------- some actions are needed on some prefixes ------------#
				elif msg == "PrefixInfosUpdated":
					prefix_infos_updated()

				#-------------------- some actions are needed on the context ------------#
				elif msg == "ContextUpdated" :
					context_updated()

				#-------------------- it is time to update duplicated address list ------------#
				elif msg == "blacklistAddrDuplUpdated" :
					black_list_AddrDupl_updated()

				#-------------------- it is time to update full cache router list ------------#
				elif msg == "blacklistRtrFullUpdated" :
					black_list_RouterFull_updated()

			except Exception as e:
				traceback.print_exc()

##########################################################################################################################
###################################################### ICMPv6 PROCESS(Echo-reply)  ##############################################
##########################################################################################################################



	def __icmpv6_process (self, process):
		'''ICMPv6 process
		
		- answers echo requests sent to this node
		'''
		self.register_message (IPv6()/ICMPv6EchoRequest(), process)

		while True:
			msg = yield
			self.debug("ICMPv6Process: Got an ICMPv6 Echo Request From:"+ repr(msg["src"]) +" , to: " + repr(msg["dst"]) )
			
			
			if msg["dst"] not in self._black_list_addrDupl :

				if msg["dst"][0] in (0xfe, 0xff): # linklocal or multicast
					src = self.__lladdr	#FIXME: this covers only the link-local scope
				else:
					src = msg["dst"]

				ereq = msg["pl"]

	
				self.__send (
					IPv6 ( src = src, dst = msg["src"])/
					ICMPv6EchoReply ( id = ereq["id"], seq = ereq["seq"], pl = ereq["pl"] )
				)



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""A module for guessing a name for the objects"""

from ttproto.core.typecheck import *
import dis, inspect, opcode, os

__all__ = [
    'NamedObject',
    'get_parent_var_name',
    'skip_parent_var_name',
]


class NamedObject:
    """A class of object whose instance name will be guessed by this module

    This function will store a name for this object in its __name__
    attribute.

    This name will be set to the value of the 'name' parameter if present.
    Otherwise the function will use the result of get_parent_var_name().

    see help(get_parent_var_name)
    """

    @typecheck
    def __init__(self, name: optional(str) = None):
        self.__name__ = name if name else get_parent_var_name()


def _print_stack(st):
    i = len(st) - 1
    for l in reversed(st):
        l4 = l[4]
        print("%d. %s() in %s:%d\n  %s" % (i, l[3], l[1], l[2], "\n" if l[4] is None else l4[0]), end=' ')
        i -= 1


_OPCODES_STORE = (
    opcode.opmap['STORE_DEREF'],
    opcode.opmap['STORE_FAST'],
    opcode.opmap['STORE_GLOBAL'],
    opcode.opmap['STORE_NAME'],
)
_OPCODE_JUMP_FORWARD = opcode.opmap['JUMP_FORWARD']
_REJECTED_NEXT_OPCODES = (opcode.opmap['BUILD_TUPLE'],
                          opcode.opmap['COMPARE_OP'],
                          opcode.opmap['LOAD_CONST'],
                          opcode.opmap['LOAD_FAST'],
                          opcode.opmap['LOAD_GLOBAL'],
                          opcode.opmap['LOAD_NAME'],
                          opcode.opmap['LOAD_ATTR'],
                          opcode.opmap['LOAD_DEREF'],
                          opcode.opmap['CALL_FUNCTION'],
                          opcode.opmap['BUILD_LIST'],
                          opcode.opmap['BUILD_MAP'],
                          opcode.opmap['ROT_TWO'],
                          opcode.opmap['BINARY_TRUE_DIVIDE'],
                          opcode.opmap['PRINT_EXPR'],
                          opcode.opmap['RETURN_VALUE']  # FIXME: maybe not reject this one
                          )


def _forward_instruction(it_instructions, offset):
    for opcode in it_instructions:
        assert opcode.offset <= offset
        if opcode.offset == offset:
            return opcode


def get_parent_var_name():
    """A function that inspects the execution stack to guess the name of
    the variable in which is stored the result of the calling function.

    NOTE: this feature might be broken at any time. The scripts should not
    rely on this feature at all. Its purpose is just to bring some useful
    indicative into the test logs.

    Examples:
        >>> blah_blah_blah = NamedObject()
        >>> blah_blah_blah.__name__
        'blah_blah_blah'

        >>> NamedObject().__name__
        '(anon)'

        >>> IPv6_multicast = IPv6(dst=IPv6Prefix("ff00::/8"))
        >>> IPv6_multicast.__name__
        'IPv6_multicast'
    """  # identify the frame of the caller of the parent function
    name = "(anon)"
    self = None
    try:
        stack = inspect.stack()
        #		_print_stack (stack)
        #		print stack[3]
        #		f = stack[3][0]
        #		print f.f_locals
        #		print dir(f)
        #		for par in 'co_argcount', 'co_cellvars', 'co_consts', 'co_filename', 'co_firstlineno', 'co_flags', 'co_freevars', 'co_lnotab', 'co_name', 'co_names', 'co_nlocals', 'co_stacksize', 'co_varnames':
        #			print "%s: %s" % (par, getattr (f.f_code, par))
        #
        #		print "-------------------------------------------------------\n"
        id = 1
        while True:
            id_next = id + 1
            layer = stack[id]
            func = layer[3]
            frame = layer[0]

            # will be set if this function is enclosed by a fake_function_skip_parent_var_name
            # from a @skip_parent_var_name
            skip_tagged = False

            try:
                if func == "type_definition":
                    name = frame.f_locals["cls"].__name__
                    raise IndexError()

                # check if this function is tagged with @typecheck
                if stack[id_next][3] == "typecheck_invocation_proxy":
                    # skip it
                    id_next += 1

                # check if this function is tagged with @skip_parent_var_name
                if stack[id_next][3] == "fake_function_skip_parent_var_name":
                    # skip it
                    id_next += 1
                    skip_tagged = True

            except IndexError:
                pass

            if not skip_tagged:
                # if the calling function is a constructor
                # then we will skip the parent callers if they are
                # constructors for the same object
                if func == '__init__' and frame.f_code.co_argcount:
                    f_self = frame.f_locals[frame.f_code.co_varnames[0]]
                    if id == 1:
                        self = f_self
                    elif not self is f_self:
                        # different object, we stop here
                        break
                else:
                    # this function is not tagged and is not a constructor
                    # -> end of recursion
                    break

            id = id_next

        #print ("frame id:", id)
        code = frame.f_code.co_code
        it_instructions = dis.get_instructions(frame.f_code)

        istr = _forward_instruction(it_instructions, frame.f_lasti)
        #print ("istr", istr)

        if istr.opname.startswith("CALL_"):
            istr = next(it_instructions)
            #print ("next", istr)
        else:
            raise Error("Current instruction is %d %s but should be a function call" % (currop, opcode.opname[currop]))
        while istr.opcode == _OPCODE_JUMP_FORWARD:
            istr = _forward_instruction(it_instructions, istr.argval)
            #print ("next", istr)
         
        nextop = istr.opcode
        #print ("nextop", opcode.opname [nextop])

        if nextop in _OPCODES_STORE:
            name = istr.argval
        else:
            if nextop not in _REJECTED_NEXT_OPCODES:
                _print_stack(stack)
                print("Now in frame", id)
                print("Warning: cannot guess the caller var name (opcode: %d %s)" % (nextop, opcode.opname[nextop]))
                global frame_fail
                frame_fail = frame

    except IndexError:
        #		print "IndexError"
        pass

    # print ("Result:", name)
    return name


def skip_parent_var_name(func):
    """A decorator for functions that should be skipped by
    get_parent_var_name() when inspecting the stack.

    Example:
        >>> def some_factory_function():
        ...   result = NamedObject()
        ...   return result
        ...
        >>> @skip_parent_var_name
        ... def some_smart_factory_function():
        ...   smart_result = NamedObject()
        ...   return smart_result
        ...
        >>> foo = some_factory_function()
        >>> foo.__name__
        'result'
        >>> bar = some_smart_factory_function()
        >>> bar.__name__
        'bar'
    """

    def fake_function_skip_parent_var_name(*k, **kw):
        return func(*k, **kw)

    return fake_function_skip_parent_var_name



from ..common import *


class TD_COAP_CORE_01(CoAPTestCase):
    """
---
TD_COAP_CORE_01:
    cfg: CoAP_CFG_BASIC
    obj: Perform GET transaction(CON mode)
    pre: Server offers the resource /test with resource content
        is not empty that handles GET with an arbitrary payload
    ref: '[COAP] 5.8.1, 1.2, 2.1, 2.2, 3.1'
    seq:
    -   s:
        - 'Client is requested to send a GET request with:'
        -   - Type = 0(CON)
            - Code = 1(GET)
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=0 and Code=1
            - "Client-generated Message ID(\u2794 CMID)"
            - "Client-generated Token(\u2794 CTOK)"
            - Uri-Path option "test"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.05(Content)
            - Message ID = CMID, Token = CTOK
            - Content-format option
            - Non-empty Payload
    -   v: Client displays the received information
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(code='get')]

    def run(self):
        self.match('client', CoAP(type='con', code='get', opt=self.uri('/test')))
        CMID = self.coap['mid']
        CTOK = self.coap['tok']

        self.next()

        if self.match('server', CoAP(code=2.05, mid=CMID, tok=CTOK, pl=Not(b'')), None):
            self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat())), 'fail')



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_11 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_11
Objective:
Perform GET transaction containing non-empty Token with a separate response (CON mode)

Configuration:
CoAP_CFG_BASIC
References:
[COAP] 2.2, 5.2.2,  5.8.1

Pre-test
conditions:
•	Server offers a resource /separate which is not served immediately and which therefore is not acknowledged in a piggybacked way.

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a GET request to server’s
resource including Token option

2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Length of the token should be between 1 to 8 Bytes
•	Uri-Path option "separate"

3
Check
Server sends acknowledgement containing:
•	Type = 2 (ACK)
•	Code = 0 (Empty)
•	Message ID = CMID
•	empty Payload


4
Check
Server sends response containing:
•	Type  = 0 (CON)
•	Code = 69 (2.05 content)
•	Server-generated Message ID (➔ SMID)
•	Token = CTOK
•	Non-empty Payload

5
Check
Client sends acknowledgement containing:
•	Type = 2 (ACK)
•	Code = 0 (Empty)
•	Message ID = SMID
•	Empty Payload


6
Verify
Client displays the response
"""

    def run (self):
        self.match_coap ("client", CoAP (	type="con",
                            code = "get",
                            tok = Not (b""),
                            opt = self.uri ("/separate"),
                ))
        self.match_coap ("client", CoAP (tok = Length (bytes, (1, 8))
                ), "fail")
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next()

        # FIXME: may be out-of-order
        if not self.match_coap	("server", CoAP (type="ack", code=0, mid=CMID,pl=b"")):
            raise self.Stop()

        self.next()

         # FIXME: this is in a different conversation
        self.match_coap ("server", CoAP (type="con", code=2.05))
        self.match_coap ("server", CoAP (
                        pl = Not (b''),
                        tok= CTOK,
                        )
                , "fail")
        SMID = self.frame.coap["mid"]

        self.next()

        self.match_coap ("client", CoAP (type="ack", code=0, mid=SMID,pl=b""))





from ..common import *


class TD_COAP_CORE_02 (CoAPTestCase):
    """
---
TD_COAP_CORE_02:
    cfg: CoAP_CFG_BASIC
    obj: Perform DELETE transaction (CON mode)
    pre: Server offers a /test resource that handles DELETE
    ref: '[COAP] 5.8.4, 1.2, 2.1, 2.2, 3.1'
    seq:
    -   s:
        - 'Client is requested to send a DELETE request with:'
        -   - Type = 0 (CON)
            - Code = 4 (DELETE)
    -   c:
        - 'The request sent by the client contains:'
        -   - Type=0 and Code=4
            - "Client-generated Message ID (\u2794 CMID)"
            - "Client-generated Token (\u2794 CTOK)"
            - Uri-Path option "test"
    -   c:
        - 'Server sends response containing:'
        -   - Code = 2.02 (Deleted)
            - Message ID = CMID, Token = CTOK
            - Content-format option if payload non-empty
            - Empty or non-empty Payload
    -   v: Client displays the received information
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(code='delete')]

    def run(self):
        self.match(
            'client',
            CoAP(type='con', code='delete', opt=self.uri('/test'))
        )
        CMID = self.coap['mid']
        CTOK = self.coap['tok']

        self.next()

        self.match(
            'server',
            CoAP(code=2.02, mid=CMID, tok=CTOK)
        )

        if self.match('server', CoAP(pl=Not(b'')), None):
            self.match(
                'server',
                CoAP(opt=Opt(CoAPOptionContentFormat())),
                'fail'
            )



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_31 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_31
Objective:
Perform CoAP Ping (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 4.3

Pre-test
conditions:
(Should work with any CoAP server)

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a "Ping" request with:
•	Type = 0 (CON)
•	Code = 0 (empty)

2
Check
The request sent by the client is four bytes and contains:
•	Type=0 and Code=0
•	Client-generated Message ID (➔ CMID)
•	Zero-length Token
•	No payload

3
Check
Server sends four-byte RST response containing:
•	Type=3 and Code=0
•	Message ID = CMID
•	Zero-length Token
•	No payload

4
Verify 	Client displays that the "Ping" was successful
    """

    def run (self):
        self.match_coap ("client", CoAP (type="con", code = 0,tok=b"",pl=b""))
        CMID = self.frame.coap["mid"]

        self.next_skip_ack()

        if self.match_coap ("server", CoAP (type=3)):
            self.match_coap ("server", CoAP (
                        code=0,
                        tok=b"",
                        pl=b"",
                    ), "fail")




#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ttproto.core.typecheck import *


class CaptureReader:
    """
    The CaptureReader interface class which will allow us to read capture files
    """

    @typecheck
    def __init__(self, file: str):
        """
        Initialize the reader with the corresponding file

        :param file: The path to the file to read
        :type file: str
        """
        raise NotImplementedError()

    def __iter__(self):
        """
        The only thing that the reader need is that it sould be iterable. Each
        returned element should be a tuple of
            - Timestamp represented as a float
            - The Message object associated to the frame
            - An Exception if one occured, None if everything went fine
        """
        raise NotImplementedError()



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import threading
from    contextlib import contextmanager

from    ttproto.core.data import Value
from    ttproto.core.union import *
from    ttproto.core.packet import *
from    ttproto.core.exceptions import Error
from    ttproto.core.typecheck import *
from    ttproto.core.lib.inet.meta import *
from    ttproto.core.lib.inet.basics import *
from    ttproto.core.lib.inet.ipv6 import *
from    ttproto.core.lib.inet.udp import *
from    ttproto.core.lib.inet.sixlowpan import *
from    ttproto.core.lib.inet.sixlowpan import sixlowpan_dispatch_bidict

# TODO: some factorisation of the compress() & decompress() functions
# FIXME: do not compress beyond the first frame (127 bytes) in case of fragmentation

__all__ = [
    "SixLowpanIPHC",
    "SixLowpanNHC",
    "SixLowpanNHC_IPExt",
    "SixLowpanNHC_UDP",
]


class _TFPad(
    metaclass=UnionClass,
    types=(UInt2, UInt4),
):
    pass


class _CompressedAddress(
    metaclass=UnionClass,
    types=(Omit, Bytes1, Bytes2, Bytes4, Bytes6, Bytes8, IPv6Address),
):
    __length_map = {
        0: Omit,
        1: Bytes1,
        2: Bytes2,
        4: Bytes4,
        6: Bytes6,
        8: Bytes8,
        16: IPv6Address,
    }

    def __new__(cls, value):
        if isinstance(value, bytes):
            assert len(value) in cls.__length_map

            return Omit() if value == b"" else cls.__length_map[len(value)](value)
        else:
            return IPv6Address(value)


class _CompressedPort(
    metaclass=UnionClass,
    types=(UInt4, UInt8, UInt16),
):
    pass


# NextHeader -> SixLowpanNHC class
sixlowpan_nhc_compress_dict = {}

# NHC_ID -> SixLowpanNHC class
sixlowpan_nhc_decompress_dict = {}


class SixLowpanNHC(
    metaclass=InetPacketClass,
    fields=[
        ("NHC_ID", "id", UInt8, InetType(None)),
        ("Value", "val", Value),
    ]):
    """
    defined in [6lowpan-hc-15]

               +----------------+---------------------------
               | var-len NHC ID | compressed next header...
               +----------------+---------------------------

                      Figure 12: LOWPAN_NHC Encoding
    """

    @classmethod
    @typecheck
    def compress(cls, next_header: int, payload: Value) -> (is_flat_value, is_binary, int):

        assert cls == SixLowpanNHC  # compress() must be reimplemented in derived classes

        try:
            new_class = sixlowpan_nhc_compress_dict[next_header]
        except KeyError:
            # cannot compress
            return Omit(), b"", 0

        return new_class.compress(next_header, payload)

    @classmethod
    @typecheck
    def decompress(cls, bin_slice: BinarySlice) -> (is_flat_value, BinarySlice, is_binary, int):
        """returns:  (decoded header, new slice, decompressed header, next header value)"""

        assert cls == SixLowpanNHC  # decompress() must be reimplemented in derived classes

        nhc_id = bin_slice[0]
        try:
            new_class = sixlowpan_nhc_decompress_dict[nhc_id]
        except KeyError:
            # cannot decompress
            raise Error("Unable to decompress 6lowpan NHC id %d" % nhc_id)

        return new_class.decompress(bin_slice)


sixlowpan_nhc_id_bidict = BidictValueType(0, SixLowpanNHC)
SixLowpanNHC.get_field(0).tag._set_bidict(sixlowpan_nhc_id_bidict)


class SixLowpanIPHC(
    metaclass=InetPacketClass,
    variant_of=SixLowpan,
    prune=0,
    fields=[
        # Base Format
        ("Dispatch", "dp", Bin(UInt3), 0b011),
        ("TF", "tf", Bin(UInt2), 0),
        ("NH", "nh", bool, 0),
        ("HLIM", "hl", Bin(UInt2), 0b10),
        ("CID", "cid", bool, 0),
        ("SAC", "sac", bool, 0),
        ("SAM", "sam", Bin(UInt2), 0),
        ("M", "m", bool, 0),
        ("DAC", "dac", bool, 0),
        ("DAM", "dam", Bin(UInt2), 0),

        # CID Extension
        ("SCI", "sci", Optional(UInt4), Omit()),
        ("DCI", "dci", Optional(UInt4), Omit()),

        # TF fields
        ("InlineECN", "iecn", Optional(Bin(UInt2)), Omit()),
        ("InlineDSCP", "idscp", Optional(Hex(UInt6)), Omit()),
        ("InlineTFPad", "ipad", Optional(_TFPad), Omit()),
        ("InlineFL", "ifl", Optional(Hex(UInt20)), Omit()),

        # NH field
        ("InlineNH", "inh", Optional(UInt8), Omit()),

        # HLIM field
        ("InlineHLIM", "ihl", Optional(UInt8), Omit()),

        # Source & Destination
        ("InlineSourceAddress", "isrc", _CompressedAddress, Omit()),
        ("InlineDestinationAddress", "idst", _CompressedAddress, Omit()),

        # Next header
        ("CompressedNextHeader", "nhc", Optional(SixLowpanNHC), Omit()),

        # TODO: add a field to contain the compressed payload ? (to allow forcing an arbitrary value)
        ("Payload", "pl", Optional(Value), Omit()),
    ],
    descriptions={
        "TF": {
            0b00: "ECN + DSCP + 4-bit Pad + Flow Label",
            0b01: "ECN + 2-bit Pad + Flow Label",
            0b10: "ECN + DSCP",
            0b11: "Elided",
        },
        "NH": {
            False: "Inline",
            True: "Compressed",
        },
        "HLIM": {
            0b00: "Hop Limit Inline",
            0b01: "Compressed hop limit = 1",
            0b10: "Compressed hop limit = 64",
            0b11: "Compressed hop limit = 255",
        },
        "CID": {
            False: "No additional context",
            True: "Context extension present",
        },
        "SAC": {
            False: "Stateless",
            True: "Stateful",
        },
        "SAM": {
            0b00: "Inline / Unspecified",
            0b01: "64 bits",
            0b10: "16 bits",
            0b11: "0 bits",
        },
        "M": {
            False: "Not Multicast",
            True: "Multicast",
        },
        "DAC": {
            False: "Stateless",
            True: "Stateful",
        },
        "DAM": {
            0b00: "Inline",
            0b01: "64 bits (multicast: 48)",
            0b10: "16 bits (multicast: 16)",
            0b11: "0 bits  (multicast: 8)",
        },
    }):
    """
    defined in [6lowpan-hc-15]:

           0                                       1
           0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
         | 0 | 1 | 1 |  TF   |NH | HLIM  |CID|SAC|  SAM  | M |DAC|  DAM  |
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+


                Figure 2: LOWPAN_IPHC base Encoding

       TF: Traffic Class, Flow Label:  As specified in [RFC3168], the 8-bit
          IPv6 Traffic Class field is split into two fields: 2-bit Explicit
          Congestion Notification (ECN) and 6-bit Differentiated Services
          Code Point (DSCP).
          00:  ECN + DSCP + 4-bit Pad + Flow Label (4 bytes)
          01:  ECN + 2-bit Pad + Flow Label (3 bytes), DSCP is elided
          10:  ECN + DSCP (1 byte), Flow Label is elided
          11:  Traffic Class and Flow Label are elided.
       NH: Next Header:
          0: Full 8 bits for Next Header are carried in-line.
          1: The Next Header field is compressed and the next header is
         encoded using LOWPAN_NHC, which is discussed in Section 4.1.

       HLIM: Hop Limit:
          00:  The Hop Limit field is carried in-line.
          01:  The Hop Limit field is compressed and the hop limit is 1.
          10:  The Hop Limit field is compressed and the hop limit is 64.
          11:  The Hop Limit field is compressed and the hop limit is 255.

       CID: Context Identifier Extension:
          0: No additional 8-bit Context Identifier Extension is used.  If
         context-based compression is specified in either SAC or DAC,
         context 0 is used.
          1: An additional 8-bit Context Identifier Extension field
         immediately follows the DAM field.

       SAC: Source Address Compression
          0: Source address compression uses stateless compression.
          1: Source address compression uses stateful, context-based
         compression.

       SAM: Source Address Mode:
          If SAC=0:
         00:  128 bits.  The full address is carried in-line.
         01:  64 bits.  The first 64-bits of the address are elided.
            The value of those bits is the link-local prefix padded with
            zeros.  The remaining 64 bits are carried in-line.
         10:  16 bits.  The first 112 bits of the address are elided.
            The value of the first 64 bits is the link-local prefix
            padded with zeros.  The following 64 bits are 0000:00ff:
            fe00:XXXX, where XXXX are the 16 bits carried in-line.
         11:  0 bits.  The address is fully elided.  The first 64 bits
            of the address are the link-local prefix padded with zeros.
            The remaining 64 bits are computed from the encapsulating
            header (e.g. 802.15.4 or IPv6 source address) as specified
            in Section 3.2.2.
          If SAC=1:
         00:  The UNSPECIFIED address, ::
         01:  64 bits.  The address is derived using context information
            and the 64 bits carried in-line.  Bits covered by context
            information are always used.  Any IID bits not covered by
            context information are taken directly from the
            corresponding bits carried in-line.  Any remaining bits are
            zero.
    """

    def describe(self, desc):
        return self.describe_payload(desc)

    def _build_message(self, ip6: optional(is_flat_value) = None) -> either((is_flat_value, is_binary, int),
                                                                            (is_flat_value, is_binary)):
        """the last value returned is the number of bytes that have been compressed in the IPv6 payload"""

        is_first_header = ip6 is None
        if is_first_header:
            # first header
            payload, payload_bin = self["pl"].build_message()
            ip6 = payload
        else:
            # subsequent header (will not include "payload") into the result
            assert self["Payload"] is None  # would be ignored anyway
            payload, payload_bin = Omit(), b""

        assert type(ip6) in (IPv6, Omit)  # TODO: support other types (would need to decode the binary)

        skip = 40

        values = self.DataList(self._fill_default_values())

        # we perform the compression only if there is an IPv6 header given
        # (otherwise we will just use the default values)
        if isinstance(ip6, IPv6):
            hwsrc, hwdst = self.get_current_iid_context()

            #######################################################
            # Traffic Class and Flow Label
            tc = ip6["tc"]
            ecn = (tc & 0b11000000) >> 6
            dscp = tc & 0b00111111
            tf = self["tf"]
            if tf is None:
                if ip6["fl"]:
                    if dscp:
                        # all are present
                        tf = 0
                    else:
                        # DSCP elided
                        tf = 1
                elif tc:
                    # Flow Label elided
                    tf = 2
                else:
                    # all are elided
                    tf = 3
                values["tf"] = tf
            if self["iecn"] is None:
                values["iecn"] = ecn if tf < 3 else Omit()
            if self["idscp"] is None:
                values["idscp"] = dscp if tf % 2 == 0 else Omit()
            if self["ipad"] is None:
                values["ipad"] = Omit() if tf >= 2 else UInt4(0) if tf == 0 else UInt2(0)
            if self["ifl"] is None:
                values["ifl"] = ip6["fl"] if tf < 2 else Omit()

            #######################################################
            # Hop Limit
            hl = ip6["hl"]
            # pre-compute a candidate value for hlim
            if self["hl"] is None:
                if hl == 1:
                    hlim = 0b01
                elif hl == 64:
                    hlim = 0b10  # FIXME: possible issue if the packet is forwarded
                elif hl == 255:
                    hlim = 0b11  # FIXME: idem, except if dsl is in fe80::/64
                else:
                    hlim = 0b00
            else:
                hlim = self["hl"]

            if self["hl"] is None:
                values["hl"] = hlim
            if self["ihl"] is None:
                values["ihl"] = Omit() if hlim else hl

            #######################################################
            # Unicast addresses compression
            def compress_unicast(ipv6_addr, hw, am_index, ac_index, addr_index):
                """returns (addr_mode, addr_comp, context_info)"""

                ci = None
                if ipv6_addr == IPV6_UNSPECIFIED_ADDRESS and am_index == "sam":  # not applicable to the dst addr
                    # special case: unspecified addr
                    ac = 1
                    am = 0
                else:
                    # interface id compression
                    if ipv6_addr[8:] == hw:
                        # full compression (Eui64 address)
                        am = 3
                    elif ipv6_addr[8:14] == b"\0\0\0\xff\xfe\0":
                        # 16-bit address pattern
                        if ipv6_addr[14:] == hw:
                            # full compression (short address)
                            am = 3
                        else:
                            # 16 bits compression
                            am = 2
                    else:
                        # no iid compression
                        am = 1

                    # prefix compression
                    if ipv6_addr[:8] == b"\xfe\x80\0\0\0\0\0\0":
                        # link-local
                        ac = 0
                    else:
                        ci = self.find_context(ipv6_addr)
                        if ci is not None and ipv6_addr != IPV6_UNSPECIFIED_ADDRESS:
                            # stateful compression
                            ac = 1
                        else:
                            # context not found -> disable all compression
                            ac = 0
                            am = 0
                            ci = None

                if self[am_index] is not None:
                    sam = self[am_index]

                if self[ac_index] is None:
                    values[ac_index] = ac
                if self[am_index] is None:
                    values[am_index] = am
                if self[addr_index] is None:
                    pfx = b"" if ac or am else ipv6_addr[:8]
                    if am == 2:
                        # 16 bits
                        iid = ipv6_addr[-2:]
                    elif am == 3:
                        # 0 bits
                        iid = b""
                    else:
                        if am == 0 and ac == 1:
                            # unspecified address
                            iid = b""
                        else:
                            # 64 bits
                            iid = ipv6_addr[-8:]

                    values[addr_index] = pfx + iid

                return ci

            #######################################################
            # Source Address
            src = ip6["src"]
            sci = compress_unicast(src, hwsrc, "sam", "sac", "isrc")

            #######################################################
            # Destination Address
            dst = ip6["dst"]
            m = dst[0] == 0xff
            if self["m"] is None:
                values["m"] = m
            if not m:
                dci = compress_unicast(ip6["dst"], hwdst, "dam", "dac", "idst")
            else:
                # multicast address compression
                dci = None
                dac = 0
                if dst[:15] == b"\xff\2\0\0\0\0\0\0\0\0\0\0\0\0\0":
                    # 8 bits
                    dam = 3
                elif dst[2:13] == b"\0\0\0\0\0\0\0\0\0\0\0":
                    # 32 bits
                    dam = 2
                elif dst[2:11] == b"\0\0\0\0\0\0\0\0\0":
                    # 48 bits
                    dam = 1
                else:
                    # 128 bits or stateful
                    dam = 0
                    dci = self.find_context(IPv6Address(dst[4:12] + bytes(8)), dst[3])
                    if dci is not None:
                        # stateful compression
                        dac = 1

                if self["dac"] is None:
                    values["dac"] = dac
                if self["dam"] is None:
                    values["dam"] = dam
                if self["idst"] is None:
                    if dac == 0:
                        if dam == 0:
                            # 128 bits
                            iid = dst
                        elif dam == 1:
                            # 48 bits
                            iid = dst[1:2] + dst[11:16]
                        elif dam == 2:
                            # 32 bits
                            iid = dst[1:2] + dst[13:16]
                        else:
                            # 8 bits
                            iid = dst[15:16]

                        values["idst"] = iid
                    else:
                        # stateful
                        assert dam == 0  # other cases are reserved

                        values["idst"] = dst[1:3] + dst[12:]

            #######################################################
            # Context Information Extension
            cid = self["cid"]
            if cid is None:
                cid = sci is not None or dci is not None
            sci = sci if sci else 0
            dci = dci if dci else 0

            if self["cid"] is None:
                values["cid"] = cid
            if self["sci"] is None:
                values["sci"] = sci if cid else Omit()
            if self["dci"] is None:
                values["dci"] = dci if cid else Omit()

            #######################################################
            # NextHeader
            if self["nhc"] is None:
                # enter and iid context because we may use to compressed addresses
                with self.encapsulating_iid_context(src, dst):
                    nhc, nhc_bin, skip_add = SixLowpanNHC.compress(ip6["nh"], ip6["pl"])
                    skip += skip_add
            else:
                nhc, nhc_bin = self["nhc"].build_message()
                skip += nhc.calc_skip()

            nh = nhc != Omit()  # TODO: document that this value may be computed from self["nhc"] (contrary to all other fields)

            if self["nh"] is None:
                values["nh"] = nh
            if self["inh"] is None:
                values["inh"] = Omit() if nh else ip6["nh"]

            if self["nhc"] is not None:
                nhc = values["nhc"]
        else:
            # build the NHC header
            nhc, nhc_bin = values["nhc"].build_message()

        #######################################################
        # Generation of the message

        # remove the payload & nhc (because they are already built)
        values.pop()
        values.pop()

        # encode the fields
        values[:], bins = zip(*(f.tag.build_message(v, None) for f, v in zip(self.fields(), values)))

        # restore the nhc & payload
        # the first header carries the whole payload (minus the headers that have been compressed)
        # (subsequent headers won't carry any payload)
        values.append(nhc)
        values.append(payload)
        bins = bins + (nhc_bin, payload_bin[skip:],)

        # assemble the result
        v = type(self)(*values)
        b = concatenate(bins)

        return (v, b) if is_first_header else (v, b, skip)

    @classmethod
    def _decode_message(cls, bin_slice, root_header=True):

        hwsrc, hwdst = cls.get_current_iid_context()

        def decoder_func():
            nonlocal bin_slice
            v = None
            for f in cls.fields():
                t = yield v
                v, bin_slice = f.tag.decode_message(t if t else f.type, bin_slice, None)
                values.append(v)
            yield v

        values = cls.DataList()

        # prepare the decoder
        decoder = decoder_func()
        next(decoder)

        def decode():
            return next(decoder)

        def decode_if(cond):
            return decoder.send(None if cond else Omit)

        def decode_as(type):
            return decoder.send(type)

        # decode the base format
        dp = decode()
        tf = decode()
        nh = decode()
        hl = decode()
        cid = decode()
        sac = decode()
        sam = decode()
        m = decode()
        dac = decode()
        dam = decode()

        # CID extension
        sci = decode_if(cid)
        dci = decode_if(cid)

        # TF fields
        iecn = decode_if(tf != 3)
        idscp = decode_if(not (tf & 1))
        ipad = decode_as(Omit if tf & 2  else (UInt2 if tf else UInt4))
        ifl = decode_if(not (tf & 2))

        # NH field
        inh = decode_if(nh == 0)

        # HLIM field
        ihl = decode_if(hl == 0)

        # src & dst addr
        def decode_unicast(am, ac):
            if am == 0:
                return decode_as(Omit if ac else IPv6Address)
            else:
                return decode_as((None, Bytes8, Bytes2, Omit)[am])

        isrc = decode_unicast(sam, sac)
        if not m:
            idst = decode_unicast(dam, dac)
        else:
            # multicast
            if dac:
                idst = decode_as(Omit if dam else Bytes6)
            else:
                idst = decode_as((IPv6Address, Bytes6, Bytes4, Bytes1)[dam])

        # compute the addresses
        def compute_unicast(am, ac, inl, hw, ci):
            if am == 0:
                return IPV6_UNSPECIFIED_ADDRESS if ac else inl
            else:
                if ac == 0:
                    pfx = b"\xfe\x80\0\0\0\0\0\0"
                else:
                    if ci == Omit():
                        raise Error("6Lowpan decompression error: packet uses stateful compression whereas CID == 0")
                    pfx, length = cls.get_context(ci)
                    pfx = pfx[:8]

                if am == 1:
                    iid = inl
                elif am == 2:
                    iid = b"\0\0\0\xff\xfe\0" + inl
                else:
                    iid = hw

                return IPv6Address(pfx + iid)

        def compute_multicast():
            if dac:
                # Stateful
                if dam == 0:
                    pfx, length = cls.get_context(dci)
                    return IPv6Address(b"".join((b"\xff", idst[:2], bytes((length,)), pfx[:8], idst[2:])))
                else:
                    return IPV6_UNSPECIFIED_ADDRESS
            else:
                # Stateless
                if dam == 0:
                    return idst
                elif dam == 1:
                    return IPv6Address(b"".join((b"\xff", idst[:1], b"\0\0\0\0\0\0\0\0\0", idst[1:])))
                elif dam == 2:
                    return IPv6Address(b"".join((b"\xff", idst[:1], b"\0\0\0\0\0\0\0\0\0\0\0", idst[1:])))
                else:
                    return IPv6Address(b"\xff\2\0\0\0\0\0\0\0\0\0\0\0\0\0" + idst)

        # decompress the addresses
        src = compute_unicast(sam, sac, isrc, hwsrc, sci)
        dst = compute_multicast() if m else compute_unicast(dam, dac, idst, hwdst, dci)

        # next compressed header
        if not nh:
            nhc = Omit()
            nhc_bin = b""
            next_header = inh
        else:
            # here were enter a pseudo_addr context because we may need to compute an IPv6 checksum
            # and enter and iid context because we may have compressed addresses
            with InetPacketValue.ipv6_pseudo_addresses_context((src, dst)), \
                 cls.encapsulating_iid_context(src, dst):

                nhc, bin_slice, nhc_bin, next_header = SixLowpanNHC.decompress(bin_slice)
        values.append(nhc)

        def value(value, default_if_omit):
            return default_if_omit if isinstance(value, Omit) else value

        # Assemble the IPv6 header
        ipv6_bin = IPv6(
            tc=(value(iecn, 0) << 6) | value(idscp, 0),
            fl=value(ifl, 0),
            len=len(bin_slice) + len(nhc_bin),
            nh=next_header,
            hl=(ihl, 1, 64, 255)[hl],
            src=src,
            dst=dst,
            pl="",
        ).build_message()[1]

        if root_header:
            # Generate the new slice
            inner_slice = BinarySlice(concatenate((ipv6_bin, nhc_bin, bin_slice.as_binary())))

            # Decode the payload
            payload, inner_slice = IPv6.decode_message(inner_slice)
            values.append(payload)

            if inner_slice:
                raise Error("Buffer not fully decoded (%d remaining bits)" % inner_slice.get_bit_length())

            return cls(*values), bin_slice.shift_bits(bin_slice.get_bit_length())
        else:
            values.append(Omit())
            return cls(*values), bin_slice, ipv6_bin + nhc_bin

    __local = threading.local()

    @staticmethod
    @contextmanager
    def encapsulating_iid_context(src, dst):
        def conv_iid(addr):
            if isinstance(addr, IPv6Address):
                # IPv6 address
                return addr[8:]
            elif isinstance(addr, Eui64Address):
                # EUI-64 address
                return bytes((addr[0] ^ 2,)) + addr[1:]
            else:
                # 802.15.4 short address
                assert isinstance(addr, bytes)
                assert len(addr) == 2
                return b"\0\0\0\xff\xfe\0" + addr

        l = SixLowpanIPHC.__local
        backup = l.iids if hasattr(l, "iids") else (None, None)

        l.iids = conv_iid(src), conv_iid(dst)
        try:
            yield
        finally:
            l.iids = backup

    @staticmethod
    def get_current_iid_context():
        l = SixLowpanIPHC.__local
        return l.iids if hasattr(l, "iids") else (None, None)

    __contextes_lock = threading.Lock()
    __contextes_stack = []
    __contextes = tuple((IPV6_UNSPECIFIED_ADDRESS, 128) for i in range(0, 16))

    @classmethod
    @contextmanager
    def contextes(cls, *k):
        """k is a list of tuples (id, prefix, length)

            eg:	((1,  "aaaa::", 64),
                 (12,b"1234560000000000", 48))

        """

        with cls.__contextes_lock:
            # copy the current context
            contextes = list(cls.__contextes)
            # push it into the stack
            cls.__contextes_stack.append(cls.__contextes)
            # update the new context
            for id, pfx, length in k:
                assert 0 <= id <= 15
                assert 0 <= length <= 128
                contextes[id] = store_data(pfx, IPv6Address), length
            # use it as the current context
            cls.__contextes = contextes

        try:
            yield
        finally:
            with cls.__contextes_lock:
                if cls.__contextes is contextes:
                    # restore the previous from the stack
                    cls.__contextes = cls.__contextes_stack.pop()
                else:
                    # may happen if the contextes are modified by multiple threads
                    # -> remove the corresponding entry in the stack
                    cls.__contextes_stack.remove(contextes)

    @classmethod
    @typecheck
    def get_context(cls, id: int) -> (IPv6Address, id):
        assert 0 <= id <= 15
        with cls.__contextes_lock:
            return cls.__contextes[id]

    @classmethod
    @typecheck
    def find_context(cls, addr: IPv6Address, length=None) -> optional(int):
        with cls.__contextes_lock:
            if length is None:
                # address lookup
                i = 0
                for pfx, length in cls.__contextes:
                    if IPv6Prefix(pfx, length).match(addr) \
                            and IPv6Prefix(addr, length).get_address() == IPv6Prefix(addr, 64).get_address():
                        return i
                    i += 1
                return None
            else:
                # prefix lookup
                try:
                    return cls.__contextes.index((addr, length))
                except ValueError:
                    return None

    @classmethod
    def print_contextes(cls):
        print("IPHC contextes")
        for i, ctx in zip(range(0, 16), cls.__contextes):
            print("  %02d -> %s/%d" % (i, ctx[0], ctx[1]))


class _IPExtNHC(
    metaclass=UnionClass,
    types=(Omit, SixLowpanNHC, SixLowpanIPHC),
):
    pass


class SixLowpanNHC_IPExt(
    metaclass=InetPacketClass,
    variant_of=SixLowpanNHC,
    prune=0,
    fields=[
        ("NHC_ID", "id", UInt4, 0b1110),
        ("ExtensionHeaderId", "eid", UInt3, 0),
        ("NextHeader", "nh", bool, False),
        ("InlineNextHeader", "inh", Optional(UInt8), Omit()),
        ("InlineLength", "ilen", Optional(UInt8), Omit()),
        ("InlineValue", "ival", Optional(bytes), Omit()),
        ("CompressedNextHeader", "nhc", _IPExtNHC, Omit()),
    ],
    descriptions={
        "nh": {
            False: "Inline",
            True: "Compressed",
        },
        "id": {
            0: "Hop-by-Hop",
            1: "Routing",
            2: "Fragment",
            3: "Destination",
            4: "Mobility",
            7: "IPv6",
        },
    }):
    __type_bidict = Bidict({
        # TODO: add extension headers

        7: IPv6,
    })

    @classmethod
    @typecheck
    def compress(cls, next_header: int, payload: Value) -> (is_flat_value, is_binary, int):

        values = cls.DataList(cls()._fill_default_values())

        if next_header == 41:  # IPv6

            # compress the next IPv6 header recursively
            iphc, iphc_bin, skip = SixLowpanIPHC()._build_message(payload)

            values["eid"] = 7
            values[:], bins = zip(*(f.tag.build_message(v, None) for f, v in zip(cls.fields(), values)))

            values["nhc"] = iphc
            bins = bins + (iphc_bin,)

            return cls(*values), concatenate(bins), skip
        else:
            assert False  # TODO: extension headers compression

    @classmethod
    @typecheck
    def decompress(cls, bin_slice: BinarySlice):

        def decoder_func():
            nonlocal bin_slice
            v = None
            for f in cls.fields():
                t = yield v
                v, bin_slice = f.tag.decode_message(t if t else f.type, bin_slice, None)
                values.append(v)
            yield v

        values = cls.DataList()

        # prepare the decoder
        decoder = decoder_func()
        next(decoder)

        def decode():
            return next(decoder)

        def decode_if(cond):
            return decoder.send(None if cond else Omit)

        def decode_as(type):
            return decoder.send(type)

        nhc_id = decode()
        eid = decode()
        nh = decode()
        if eid == 7:
            # IPv6
            decode_as(Omit())  # inh
            decode_as(Omit())  # ilen
            decode_as(Omit())  # ival

            nhc, bin_slice, ip6_bin = SixLowpanIPHC._decode_message(bin_slice, False)
            values.append(nhc)

            return cls(*values), bin_slice, ip6_bin, 41

        else:
            assert False  # TODO implement extension headers decompression


class SixLowpanNHC_UDP(
    metaclass=InetPacketClass,
    variant_of=SixLowpanNHC,
    prune=0,
    fields=[
        ("NHC_ID", "id", UInt5, 0b11110),
        ("Checksum", "c", bool, True),
        ("Ports", "p", UInt2, 0),
        ("InlineSourcePort", "isp", _CompressedPort, UInt16(0)),
        ("InlineDestinationPort", "idp", _CompressedPort, UInt16(0)),
        ("InlineChecksum", "ic", Optional(Hex(UInt16)), Omit()),
    ],
    descriptions={
        "c": {
            False: "Inline",
            True: "Elided",
        },
        "p": {
            0b00: "Inline",
            0b01: "16 bits src/8 bits dst",
            0b10: "8 bits src/16 bits dst",
            0b11: "4 bits src/4 bits dst",
        },
    }):
    def calc_skip(self):
        return 8

    @classmethod
    @typecheck
    def compress(cls, next_header: int, udp: Value) -> (is_flat_value, is_binary, int):

        assert isinstance(udp, UDP)  # TODO: support other types (would have to decode it)

        sport = udp["sport"]
        dport = udp["dport"]

        if (sport & 0xfff0) == (dport & 0xfff0) == 0xf0b0:
            # 4 bits + 4 bits
            isp = UInt4(sport & 0xf)
            idp = UInt4(dport & 0xf)
            p = 3
        elif (sport & 0xff00) == 0xf000:
            # 8 bits + 16 bits
            isp = UInt8(sport & 0xff)
            idp = dport
            p = 2
        elif (dport & 0xff00) == 0xf000:
            # 16 bits + 8 bits
            isp = sport
            idp = UInt8(dport & 0xff)
            p = 1
        else:
            # 16 bits + 16 bits
            isp = sport
            idp = dport
            p = 0

        return cls(
            c=0,  # FIXME: how to specify that the checksum should be compressed ?
            p=p,
            ic=udp["Checksum"],
            isp=isp,
            idp=idp,
        ).build_message() + (8,)

    @classmethod
    @typecheck
    def decompress(cls, bin_slice: BinarySlice):

        def decoder_func():
            nonlocal bin_slice
            v = None
            for f in cls.fields():
                t = yield v
                v, bin_slice = f.tag.decode_message(t if t else f.type, bin_slice, None)
                values.append(v)
            yield v

        values = cls.DataList()

        # prepare the decoder
        decoder = decoder_func()
        next(decoder)

        def decode():
            return next(decoder)

        def decode_if(cond):
            return decoder.send(None if cond else Omit)

        def decode_as(type):
            return decoder.send(type)

        nhc_id = decode()
        c = decode()
        p = decode()
        isp = decode_as((UInt16, UInt16, UInt8, UInt4)[p])
        idp = decode_as((UInt16, UInt8, UInt16, UInt4)[p])
        ic = decode_if(not c)

        value = cls(*values)

        # decompress the values
        def port_value(inl):
            if isinstance(inl, UInt16):
                return inl
            elif isinstance(inl, UInt8):
                return 0xf000 | inl
            else:
                assert isinstance(inl, UInt4)
                return 0xf0b0 | inl

        sport = port_value(isp)
        dport = port_value(idp)
        length = len(bin_slice) + 8

        if ic == Omit():
            # recompute the checksum on the fly
            udp_bin = UDP(sport, dport, length, None, bin_slice.raw()).build_message()[1][:8]
        else:
            # use the inline checksum
            udp_bin = UDP(sport, dport, length, ic, b"").build_message()[1]

        return value, bin_slice, udp_bin, 17


# register the NHC compression headers
sixlowpan_nhc_compress_dict.update((
    (0, SixLowpanNHC_IPExt),  # Hop by Hop
    (43, SixLowpanNHC_IPExt),  # Routing Header
    (44, SixLowpanNHC_IPExt),  # Fragment Header
    (60, SixLowpanNHC_IPExt),  # Destination Options Header
    (135, SixLowpanNHC_IPExt),  # Mobility Header
    (41, SixLowpanNHC_IPExt),  # IPv6
    (17, SixLowpanNHC_UDP),  # UDP
))
for i in range(0b11100000, 0b11110000):
    sixlowpan_nhc_decompress_dict[i] = SixLowpanNHC_IPExt
for i in range(0b11110000, 0b11111000):
    sixlowpan_nhc_decompress_dict[i] = SixLowpanNHC_UDP

sixlowpan_nhc_decompress_dict.update((
    (0, SixLowpanNHC_IPExt),  # Hop by Hop
    (43, SixLowpanNHC_IPExt),  # Routing Header
    (44, SixLowpanNHC_IPExt),  # Fragment Header
    (60, SixLowpanNHC_IPExt),  # Destination Options Header
    (135, SixLowpanNHC_IPExt),  # Mobility Header
    (41, SixLowpanNHC_IPExt),  # IPv6
    (17, SixLowpanNHC_UDP),  # UDP
))

# register the 6lowpan dispatch values "011xxxxx"
for i in range(0b01100000, 0b10000000):
    sixlowpan_dispatch_bidict[i] = SixLowpanIPHC



import unittest

from ttproto.core.analyzer import Conversation, Node
from ttproto.core.data import Value
from ttproto.core.typecheck3000 import InputParameterError
from ttproto.core.lib.all import *


class ConversationTestCase(unittest.TestCase):
    """
    Test class for the Conversation class
    """

    # #################### Tests parameters #########################

    # The nodes used here (CoAP ones)
    NODES = [Node('client', UDP(dport=5683)), Node('server', UDP(sport=5683))]

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the conversation instance
        """
        self.conversation = Conversation(self.NODES)

    # #################### Tests functions #########################

    # ##### __init__
    def test___init__single_node_not_a_list(self):
        with self.assertRaises(InputParameterError):
            self.conversation = Conversation(self.NODES[0])

    def test___init__single_node_in_a_list(self):
        with self.assertRaises(ValueError):
            self.conversation = Conversation([self.NODES[0]])

    def test___init__empty_list(self):
        with self.assertRaises(ValueError):
            self.conversation = Conversation([])

    # ##### nodes
    def test_nodes(self):

        # Get and check nodes
        nodes = self.conversation.nodes
        self.assertIsInstance(nodes, dict)
        self.assertEqual(len(nodes), 2)
        for node_name, node_value in nodes.items():
            self.assertIsInstance(node_name, str)
            self.assertIsInstance(node_value, Value)

        # Check that setting it is blocked
        with self.assertRaises(AttributeError):
            self.conversation.nodes = nodes

    # ##### __bool__
    def test___bool__(self):
        self.assertTrue(self.conversation)


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from	contextlib	import contextmanager
import	struct, threading

from	ttproto.core.typecheck	import *
from	ttproto.core.exceptions	import Error, push_location
from	ttproto.core.data	import *
from	ttproto.core.subtype	import SubtypeClass
from	ttproto.core.templates	import Range, Length
from	ttproto.core.packet	import *
from	ttproto.core.list	import *
from	ttproto.core.primitive	import IntValue

__all__ = [
	'UnsignedBigEndianIntClass',
	'FixedLengthBytesClass',
	'InetType',
	'InetLength',
	'InetCount',
	'InetVariant',
	'InetIPv6Checksum',
	'InetPaddedPrefix',
	'InetPacketValue',
	'InetPacketClass',
	'InetListValue',
	'InetOrderedListClass',
	'InetUnorderedListClass',
	'Hex',
	'Bin',
]

def UnsignedBigEndianIntClass (size):

	class UnsignedBigEndianIntValue (metaclass = SubtypeClass (Range (int, 0, 2**size))):

		@typecheck
		def _build_message (self) -> is_flatvalue_binary:

			result = []
			v = self
			if not self.__size % 8:
				# multiple of 8 bit
				for s in range (0, self.__size//8):
					result.append (v & 255)
					v >>= 8
				return self, bytes (reversed (result))
			else:
				# not a multiple of 8 bits

				# add padding bits
				v <<= (-self.__size % 8)
				for s in range (0, self.__size//8 + 1):
					result.append (v & 255)
					v >>= 8
				return self, (bytes (reversed (result)), self.__size%8)

		@classmethod
		@typecheck
		def decode_message (cls, bin_slice: BinarySlice) -> is_flatvalue_binslice:
			# decode the int
			v = 0
			nb = cls.__size // 8
			remainder = cls.__size % 8
			if not remainder:
				# multiple of 8 bit
				sl = bin_slice[:nb]
				for i in range (0, nb):
					v = (v << 8) | sl[i]
			else:
				# not a multiple of 8 bits

				# add padding bits
				sl = bin_slice.bit_slice (0, cls.__size)
				for i in range (0, nb):
					v = (v << 8) | sl[i]
				v = (v << remainder) | (sl[nb] >> (8 - remainder))

			return cls (v), bin_slice.shift_bits (cls.__size)

	def MetaclassFunc (name, bases, classdict):
		assert len (bases) == 0 # FIXME: it could be useful to have multiple inheritance in some cases ?

		bases = UnsignedBigEndianIntValue,
		classdict["_UnsignedBigEndianIntValue__size"] = size

		return type (name, bases, classdict)

	return MetaclassFunc


#TODO: remove this function
def value_metaclass (func):

	class TmpMetaclass:
		def __init__ (self, *k, **kw):
			self.__k  = k
			self.__kw = kw

		def __call__ (self, name, bases, classdict):
			new_base = func (*self.__k, **self.__kw)

			# prepend the new base to the provided bases
			bases = (new_base,) + bases

			return type (name, bases, classdict)

	return TmpMetaclass


@value_metaclass
def FixedLengthBytesClass (size: int):

	class FixedLengthBytesValue (metaclass = SubtypeClass (Length (bytes, size))):

		@typecheck
		def __new__ (cls, value = None):
			if value is None:
				value = b"\0" * size
			# TODO: check that the resulting value is valid
			result = super().__new__(cls, value)

			if len (result) != size:
				raise Error ("Invalid value length (%d instead of %d)" % (len (value), size))

			return result

		@classmethod
		def _decode_message (cls, bin_slice):

			return bin_slice[:size].raw(), bin_slice[size:]

		#TODO: support str in __new__ too
		def __str__ (self):
			return ":".join ("%02x" % c for c in self)

	return FixedLengthBytesValue



class __OptFieldRefTag (PacketValue.Tag):
	@typecheck
	def __init__ (self, field_name: optional (str) = None, default = 0):

		super().__init__ (default)
		self.field_name = field_name

	def init (self, *k):
		super().init (*k)
		self.__ref_field_id = None if self.field_name is None else self.get_packet_type().get_field_id(self.field_name)

	def get_ref_id(self):
		return self.__ref_field_id


class _RestartDecodingAs (Exception):
	def __init__ (self, type):
		self.type = type

class InetType (__OptFieldRefTag):
	@staticmethod
	@contextmanager
	def __decode_context (ctx, expected_type: type):
		with ctx.replace_attr ("expected_type", expected_type):
			yield

	@typecheck
	def __init__ (self, bidict: optional (Bidict), field_name: optional (str) = None):
		super().__init__ (field_name)
		self.__bidict = bidict

	def _set_bidict (self, bidict: Bidict):
		assert self.__bidict is None
		self.__bidict = bidict

	def compute (self, seq, values_bins):
		assert self.__bidict is not None  # must have been initialised

		v = seq if self.get_ref_id() is None else values_bins[self.get_ref_id()][0]

		return self.__bidict[:v.get_type()]

	def post_decode (self, ctx, value):
		assert self.__bidict is not None  # must have been initialised

		expected_type = self.__bidict[value]

		if self.get_ref_id() is not None:
			# set a handler -> a 'with' context to change the slice on the fly
			ctx.push_context (self.get_ref_id(), self.__decode_context (ctx, expected_type))
		else:
			# try to change the type on the fly

			# check that we can safely change the variant
			nb_decoded = len (ctx.values)
			v = expected_type
			prune = False
			while v != ctx.variant:
				if v.get_prune_id() < nb_decoded:
					prune = True
				v = v.get_base_variant()

				if v is None:
					raise Error ("The new variant must be based on the current one") # FIXME: this could be relaxed
			try:
				if not (expected_type._decode_message.__func__ is ctx.variant._decode_message.__func__):
					# the new variant has a different _decode_message() function
					# -> we must restart decoding from scratch
					raise _RestartDecodingAs (expected_type)
			except AttributeError:
				pass

			if prune:
				# we have already decode too many fields to be able to change the type on the fly
				# -> restart decoding from the beginning of the packet
				ctx.reset()
				ctx.remaining_slice	= ctx.initial_slice
				ctx.values[:]		= ()
				ctx.field_id		= -1

			# change the variant
			ctx.variant = expected_type


class InetVariant (InetType):
	def __init__ (self):
		super().__init__ (None)

	@typecheck
	def init (self, *k):
		super().init (*k)
		self._set_bidict (self.get_packet_type().get_variants_bidict())


class InetLength (__OptFieldRefTag):
	@staticmethod
	@contextmanager
	def __decode_context (ctx, value):
		# generate a shorter slice
		#TODO: detect possible errors (slice too short)
		with ctx.replace_attr ("remaining_slice", ctx.remaining_slice.bit_slice (0, value)):
			yield

			# ensure that we decoded all the bits
			bl = ctx.remaining_slice.get_bit_length()
			if bl:
				raise Error("Did not decode all the content of the slice (%d bits remaining)" % bl)

		# move forward n bits in the restored slice
		ctx.remaining_slice = ctx.remaining_slice.shift_bits (value)

	@typecheck
	def __init__ (self, field_name: optional (str) = None, unit: int = 1):
		super().__init__ (field_name)
		self.__unit = unit * 8

	def compute (self, seq, values_bins):
		if self.get_ref_id() is None:
			l = 0
			for vb in values_bins:
				l += get_binary_length (vb[1])
		else:
			l = get_binary_length (values_bins[self.get_ref_id()][1])

		if l % self.__unit:
			raise Error ("Invalid length: must be a multiple of %d bytes" % (self.__unit // 8))
		return l // self.__unit

	def post_decode (self, ctx, value):

		value *= self.__unit

		if self.get_ref_id() is None:
			assert ctx.initial_slice.same_buffer_as (ctx.remaining_slice) # we must still be working on the same slice

			right = ctx.initial_slice.get_left() + value
			if right != ctx.initial_slice.get_right():
				remaining_bits = right - ctx.remaining_slice.get_left()
				if remaining_bits < 0:
					raise Error("InetLength: expected length is to small: %d (%d bits), %d bits decoded so far" % (
						value / self.__unit,
						value,
						ctx.remaining_slice.get_left() - ctr.initial_slice.get_left()
					))

				else:
					final_slice = ctx.initial_slice.shift_bits (value)
					# TODO: update the initial_slice too ???
					ctx.remaining_slice = ctx.remaining_slice.bit_slice (0, remaining_bits)

					def cleaner():
						ctx.remaining_slice = final_slice
					ctx.push_cleaner (cleaner)
		else:
			# set a handler -> a 'with' context ! to change the slice on the fly
			ctx.push_context (self.get_ref_id(), self.__decode_context (ctx, value))

class InetCount (__OptFieldRefTag):
	@staticmethod
	@contextmanager
	def __decode_context (ctx, count):
		with ctx.replace_attr ("expected_count", count):
			yield

	@typecheck
	def __init__ (self, field_name: str):
		super().__init__ (field_name)

	def compute (self, seq, values_bins):
		assert self.get_ref_id() is not None

		return len (values_bins[self.get_ref_id()][0])

	def post_decode (self, ctx, value):
		# set a handler -> a 'with' context ! to change the slice on the fly
		ctx.push_context (self.get_ref_id(), self.__decode_context (ctx, value))


class InetIPv6Checksum (PacketValue.Tag):

	def __init__ (self):
		super().__init__ (0)

	def compute (self, seq, values_bins):
		# schedule the computation of the checksum if not set

		try:
			previous_func = seq._post_build_fields
		except AttributeError:
			def previous_func (dummy):
				pass

		def post_build_fields (values_bins):

			# call the previous _post_build_fields function
			previous_func (values_bins)

			# compute the pseudo-header
			src_dst = seq._get_ipv6_pseudo_addresses()
			if not src_dst:
				# no parent IPv6 header -> cannot compute the pseudo-header
				return

			assert len (src_dst[0]) == len(src_dst[1]) == 16 # IPv6 only for the moment

			payload = concatenate (vb[1] for vb in values_bins)
			length = len (payload)

			import ttproto.core.lib.inet.ipv6

			pseudo_msg = bytes().join ((src_dst[0], src_dst[1], struct.pack(">LHBB", length, 0, 0, ttproto.core.lib.inet.ipv6.ip_next_header_bidict[:seq.get_type()]), payload, bytes((0,) if length % 2 else ())))

			# compute the checksum
			chksum = self.checksum (pseudo_msg)

			chksum = (chksum >> 8) | ((chksum & 0xff) << 8)# FIXME: the endianness is not correct

			values_bins[self.get_field_id()] = self.get_field().store_data(chksum).build_message()

		seq._post_build_fields = post_build_fields

		return 0

	@staticmethod
	def carry_around_add (a, b):
		c = a + b
		return (c & 0xffff) + (c >> 16)

	def checksum (self,msg):
		s = 0
		for i in range(0, len(msg), 2):
			w = msg[i] + (msg[i+1] << 8)
			s = self.carry_around_add(s, w)
		return ~s & 0xffff

class InetPaddedPrefix (__OptFieldRefTag):
	@typecheck
	def __init__ (self, pad_step: one_of (16, 32, 64), pf_length_field_name: str):

		super().__init__ (pf_length_field_name, bytes(16))
		self.__step = pad_step

	def build_message (self, value, values_bins):

		v, b = value.build_message()
		assert isinstance (b, bytes) and len (b) == 16  # that's the length of the full prefix

		p_len = values_bins[self.get_ref_id()][0]

		expected_len = (p_len + self.__step - 1) // self.__step * self.__step // 8

		return v, (b[:expected_len] if len(b) > expected_len else b)

	def decode_message (self, type_, bin_slice, ctx):

		expected_len = (ctx.values[self.get_ref_id()] + self.__step - 1) // self.__step * self.__step // 8

		v = bin_slice[:expected_len].raw()
		if expected_len < 16:
			v += bytes(16 - expected_len)

		return type_ (v), bin_slice[expected_len:]


class _DecodeContext:
	def __init__ (self):
		self.__contextes = {}
		self.__cleaners = []

	def reset (self):
		self.__contextes.clear()
		self.__cleaners[:]=()

	@typecheck
	def push_context (self, field_id: int, obj):
		assert hasattr (obj, "__enter__") and hasattr (obj, "__exit__")
#			print "**push context", field_id, obj
		if field_id not in self.__contextes:
			self.__contextes[field_id] = [obj]
		else:
			self.__contextes[field_id].append (obj)

	def pop_context (self, field_id):

		if field_id in self.__contextes:
			l = self.__contextes [field_id]
			if l:
#					print "**pop context", field_id, l[0]
				return l.pop(0)
		return None

	@contextmanager
	def replace_attr (self, name, value):
		backup = getattr (self, name)
		setattr (self, name, value)
		try:
			yield
		finally:
			setattr (self, name, backup)

	def push_cleaner (self, func):
		self.__cleaners.insert(0, func)

	def leave (self):
		for c in self.__cleaners:
			c()

class InetPacketValue (PacketValue):


	### class methods

	@classmethod
	def metaclass_func (cls, *k, variant_of: optional (type) = None, **kw):
		assert variant_of is None or issubclass (variant_of, InetPacketValue)

		result = super().metaclass_func (*k, variant_of = variant_of, **kw)

		i=0
		for f in result.fields():
			if isinstance (f.tag, InetVariant):
				result.__cls_variant_field_id = i
				break
			i+=1
		else:
			result.__cls_variant_field_id = None

		return result

	@classmethod
	def get_variant_field_id (cls):
		return cls.__cls_variant_field_id


	### instance methods

	@typecheck
	def _build_message (self) -> is_flatvalue_binary:

		values_bins = [(v, b"") for v in self._fill_default_values()]
		values_bins = [f.tag.build_message(v_b[0], values_bins) for v_b, f in zip (values_bins, self.fields())]

		i = -1
		for field in self.fields():
			i += 1
			tag = field.tag
			if self[i] is not None or not hasattr(tag, "compute"):
				continue

			values_bins[i] = tag.build_message (field.store_data (tag.compute (self, values_bins), none_is_allowed = False), values_bins)

		if hasattr (self, "_post_build_fields"): #TODO: use a context instead
			self._post_build_fields (values_bins)
			del self._post_build_fields

		values, bins = zip (*values_bins)

		return type(self) (*values), concatenate (bins)

	@typecheck
	def _build_message_ipv6_header (self) -> is_flatvalue_binary:
		"""Specialised method for building an IPv6 heades

		Its purpose is to collect the source/destination addresses
		needed to assemble the pseudo-header when computing a checksum.
		"""

		# index of the src & dst address fields
		id_src = 6
		id_dst = 7

		# fill the values if we dont have any
		# source address
		src, dst = self[6], self[7]

		if src is None:
			src = self.get_field(6).tag.get_default_value()
		if dst is None:
			dst = self.get_field(7).tag.get_default_value()

		with self.ipv6_pseudo_addresses_context ((src, dst)):
			return InetPacketValue._build_message (self)

	__local = threading.local()

	@staticmethod
	@contextmanager
	def ipv6_pseudo_addresses_context (src_dst):
		l = InetPacketValue.__local
		backup = l.pseudo_ipv6_addresses if hasattr(l, "pseudo_ipv6_addresses") else None

		l.pseudo_ipv6_addresses = src_dst
		try:
			yield
		finally:
			l.pseudo_ipv6_addresses = backup

	@staticmethod
	def _get_ipv6_pseudo_addresses():
		l = InetPacketValue.__local
		return l.pseudo_ipv6_addresses if hasattr(l, "pseudo_ipv6_addresses") else None

	@classmethod
	@typecheck
	def __decode_field (cls, ctx: _DecodeContext, field_id: int):

		obj = ctx.pop_context (field_id)
		if obj:
			with obj:
				cls.__decode_field (ctx, field_id)
		else:
			field = ctx.variant.get_field (field_id)

			expected_type = ctx.expected_type if ctx.expected_type else field.type

			assert issubclass (expected_type, field.type) or (field.optional and issubclass (expected_type, Omit))# FIXME: should be checked earlier

			if ctx.expected_count is not None:
				assert ctx.expected_count >= 0

				# Here we generate a wrapper class for the expected type, that will passe the 'count'
				# when decoding the message
				# TODO: find a more generic way to pass arbitraty parameters to the decode function
				#	This is a very ugly hack: we define an inherited class for expected_type
				#	and use a metaclass to overload 'isinstance()' so that Value.decode_message()
				#	does not complain about mismatching types
				expected_type_orig = expected_type
				@classmethod
				def _decode_message (cls, bin_slice):
					return expected_type_orig._decode_message (bin_slice, count = ctx.expected_count)
				class metaclass (type):
					def __subclasscheck__ (self, instance):
						return issubclass (instance, expected_type_orig)
					def __instancecheck__ (self, instance):
						return isinstance (instance, expected_type_orig)
				expected_type = metaclass (expected_type.__name__, (expected_type,), {"_decode_message": _decode_message})

			value, ctx.remaining_slice = field.tag.decode_message (expected_type, ctx.remaining_slice, ctx)

			ctx.values.append (value)

			if hasattr (field.tag, "post_decode"):
				field.tag.post_decode (ctx, value)

#			print "decoded", field.type, repr(value)


	@classmethod
	def _decode_message (cls, bin_slice):

		ctx = _DecodeContext ()
		ctx.expected_type	= None
		ctx.expected_count	= None
		ctx.initial_slice	= bin_slice
		ctx.remaining_slice	= bin_slice
		ctx.values		= []
		ctx.cls			= cls
		ctx.variant		= cls
		ctx.field_id		= 0

		try:
			while ctx.field_id < ctx.variant.get_length():
				cls.__decode_field (ctx, ctx.field_id)

				# TODO: verify the checksum
				ctx.field_id += 1

			ctx.leave()

		except _RestartDecodingAs as e: # TODO: maybe this should be generalised in Value.decode_message()
			return e.type._decode_message (bin_slice)

		except Exception as e:
			push_location (e, ctx.variant, ctx.variant.get_field (ctx.field_id).name)
			raise

		return ctx.variant (*ctx.values), ctx.remaining_slice

InetPacketClass = InetPacketValue.metaclass_func

class InetListValue (ListValue):
	### class methods

	### instance methods
	pass

InetListClass =  InetListValue.metaclass_func

def InetOrderedListClass (name, bases, classdict, content_type):
	return InetListClass (name, bases, classdict, content_type, True)

def InetUnorderedListClass (name, bases, classdict, content_type):
	return InetListClass (name, bases, classdict, content_type, False)


class Hex:
	"""Factory class to display an int in hex format"""

	__instances = {}

	@typecheck
	def __new__ (cls, t = int):
		if t == int:
			t = IntValue

		assert issubclass (t, IntValue)

		try:
			return cls.__instances[t]
		except KeyError:

			new_type = type (t.__name__, (t,), {})
			try:
				s = new_type._UnsignedBigEndianIntValue__size
			except AttributeError:
				def __str__ (self):
					return hex (self)
			else:
				def __str__ (self):
					return ("-0x" if self < 0 else "0x") + format (abs (self), self.__format)
				new_type.__format = "0%dx" % ((s + 3) / 4)

			new_type.__str__ = __str__
			new_type.__repr__ = __str__

			cls.__instances[t] = new_type
			return new_type

class Bin:
	"""Factory class to display an int in bin format"""

	__instances = {}

	@typecheck
	def __new__ (cls, t = int):
		if t == int:
			t = IntValue

		assert issubclass (t, IntValue)

		try:
			return cls.__instances[t]
		except KeyError:

			new_type = type (t.__name__, (t,), {})
			try:
				s = new_type._UnsignedBigEndianIntValue__size
			except AttributeError:
				def __str__ (self):
					return bin (self)
			else:
				def __str__ (self):
					return ("-0b" if self < 0 else "0b") + format (abs (self), self.__format)
				new_type.__format = "0%db" % s

			new_type.__str__ = __str__
			new_type.__repr__ = __str__

			cls.__instances[t] = new_type
			return new_type


